import pytest
import requests

from secedgar.cik_lookup import CIKLookup
from secedgar.client import NetworkClient
from secedgar.core import QuarterlyFilings
from secedgar.tests.utils import AsyncMockResponse, MockResponse, datapath


@pytest.fixture(scope="module")
def monkeymodule():
    from _pytest.monkeypatch import MonkeyPatch
    mpatch = MonkeyPatch()
    yield mpatch
    mpatch.undo()


@pytest.fixture(scope="session")
def monkeysession():
    from _pytest.monkeypatch import MonkeyPatch
    mpatch = MonkeyPatch()
    yield mpatch
    mpatch.undo()


@pytest.fixture(autouse=True)
def no_http_requests(request, monkeysession):
    if "smoke" in request.keywords:
        return

    def external_request_mock(object, *args, **kwargs):
        raise RuntimeError(
            f"""A request to an external source was about to be made by {object}.
            Please provide mock for {object}.""")

    to_avoid = ("requests.Session.get", "aiohttp.ClientSession.get")

    for avoid in to_avoid:
        monkeysession.setattr(avoid, external_request_mock)


@pytest.fixture(scope="session")
def mock_user_agent():
    return "Example (me@example.com)"


@pytest.fixture(scope="session")
def mock_filing_response(monkeysession):
    monkeysession.setattr(
        NetworkClient, "fetch",
        lambda *args, **kwargs: AsyncMockResponse(content=bytes(
            "Testing...", "utf-8")).read())


@pytest.fixture(scope="session")
def mock_master_idx_file(monkeysession):
    """Mock idx file from DailyFilings."""

    def _mock_master_idx_file(*args, **kwargs):
        with open(datapath("filings", "master", "master.idx")) as f:
            return f.read()

    monkeysession.setattr(QuarterlyFilings, "_get_master_idx_file",
                          _mock_master_idx_file)


@pytest.fixture(scope="session")
def mock_filing_data(monkeysession):
    """Mock data from filing."""
    monkeysession.setattr(requests.Session, "get",
                          MockResponse(content=bytes("Testing...", "utf-8")))


@pytest.fixture
def tmp_data_directory(tmpdir_factory):
    return str(tmpdir_factory.mktemp("tmp_data"))


@pytest.fixture(scope="session")
def mock_cik_validator_get_multiple_ciks(monkeysession):
    """Mocks response for getting a single CIK."""
    monkeysession.setattr(CIKLookup, "get_ciks",
                          lambda *args: {"aapl": "0000320193", "msft": "1234", "amzn": "5678"})


@pytest.fixture(scope="session")
def mock_single_cik_filing(monkeysession):
    """Returns mock response of filinghrefs for getting filing URLs."""
    monkeysession.setattr(NetworkClient, "get_response",
                          MockResponse(datapath_args=["filings", "aapl_10q_filings.xml"]))


@pytest.fixture(scope="session")
def real_test_client():
    """``NetworkClient`` to use when running live smoke tests."""
    return NetworkClient(user_agent="Example User (example@example.com)")



import functools
import warnings

import requests

from secedgar.client import NetworkClient
from secedgar.exceptions import CIKError, EDGARQueryError


@functools.lru_cache()
def get_cik_map(user_agent):
    """Get dictionary of tickers and company names to CIK numbers.

    Uses ``functools.lru_cache`` to cache response if used in later calls.
    To clear cache, use ``get_cik_map.cache_clear()``.

    .. note::
       All company names and tickers are normalized by converting to upper case.

    Returns:
        Dictionary with keys "ticker" and "title". To get dictionary
            mapping tickers to CIKs, use "ticker". To get
            company names mapped to CIKs, use "title".

    .. note::

       If any tickers or titles are ``None``, the ticker or title will
       be excluded from the returned dictionary.

    .. versionadded:: 0.1.6
    """
    headers = {'user-agent': user_agent}
    response = requests.get("https://www.sec.gov/files/company_tickers.json", headers=headers)
    json_response = response.json()
    return {key: {v[key].upper(): str(v["cik_str"]) for v in json_response.values()
                  if v[key] is not None}
            for key in ("ticker", "title")}


class CIKLookup:
    """CIK Lookup object.

    Given list of tickers/company names to lookup, this object can return associated CIKs.

    Args:
        lookup (Union[str, list]): Ticker, company name, or list of tickers and/or company names.
        client (Union[None, secedgar.client.NetworkClient]): A network client object to use.
            If ``None`` is given, then will pass kwargs to :class:`secedgar.client.NetworkClient`.
            See :class:`secedgar.client.NetworkClient` for more details. Defaults to ``None``.

    .. warning::

        If ``None`` is passed to client, you must specify your ``user_agent`` in ``kwargs``.
        For example, ``CIKLookup(lookup=["aapl"], user_agent="Name (email@example.com")``.

    .. versionadded:: 0.1.5
    """

    def __init__(self, lookups, client=None, **kwargs):
        if lookups and isinstance(lookups, str):
            self._lookups = [lookups]  # make single string into list
        else:
            # Check that iterable only contains strings and is not empty
            if not (lookups and all(type(o) is str for o in lookups)):
                raise TypeError("CIKs must be given as string or iterable.")
            self._lookups = lookups
        self._params = {}
        self._client = client or NetworkClient(**kwargs)
        self._lookup_dict = None
        self._ciks = None

    @property
    def ciks(self):
        """:obj:`list` of :obj:`str`: List of CIKs (as string of digits)."""
        if self._ciks is None:
            self._ciks = list(self.lookup_dict.values())
        return self._ciks

    @property
    def lookups(self):
        """`list` of `str` to lookup (to get CIK values)."""
        return self._lookups

    @property
    def path(self):
        """str: Path to add to client base."""
        return "cgi-bin/browse-edgar"

    @property
    def client(self):
        """``secedgar.client.NetworkClient``: Client to use to fetch requests."""
        return self._client

    @property
    def params(self):
        """:obj:`dict` Search parameters to add to client."""
        return self._params

    @property
    def lookup_dict(self):
        """:obj:`dict`: Dictionary that makes tickers and company names to CIKs."""
        if self._lookup_dict is None:
            self._lookup_dict = self.get_ciks()
        return self._lookup_dict

    # TODO: Add mock to test this functionality
    def _get_lookup_soup(self, lookup):
        """Gets `BeautifulSoup` object for lookup.

        First tries to lookup using CIK. Then falls back to company name.

        .. warning::
           Only to be used internally by `_get_cik_from_html` to get CIK from lookup.

        Args:
            lookup (str): CIK, company name, or ticker symbol to lookup.

        Returns:
            soup (bs4.BeautifulSoup): `BeautifulSoup` object to be used to get
                company CIK.
        """
        try:  # try to lookup by CIK
            self._params['CIK'] = lookup
            return self._client.get_soup(self.path, self.params)
        except EDGARQueryError:  # fallback to lookup by company name
            self.params.pop('CIK')  # delete this parameter so no conflicts arise
            self._params['company'] = lookup
            return self._client.get_soup(self.path, self.params)

    def _get_cik_from_html(self, lookup):
        """Gets CIK from `BeautifulSoup` object.

        .. warning: This method will warn when lookup returns multiple possibilities for a
            CIK are found.

        Args:
            lookup (str): CIK, company name, or ticker symbol which was looked up.

        Returns:
            CIK (str): CIK for lookup.
        """
        self._validate_lookup(lookup)
        soup = self._get_lookup_soup(lookup)
        try:  # try to get single CIK for lookup
            span = soup.find('span', {'class': 'companyName'})
            return span.find('a').getText().split()[0]  # returns single CIK
        except AttributeError:  # warn and skip if multiple possibilities for CIK found
            warning_message = """Lookup '{0}' will be skipped.
                          Found multiple companies matching '{0}':
                          {1}""".format(lookup, '\n'.join(self._get_cik_possibilities(soup)))
            warnings.warn(warning_message)
        finally:
            # Delete parameters after lookup
            self._params.pop('company', None)
            self._params.pop('CIK', None)

    @staticmethod
    def _get_cik_possibilities(soup):
        """Get all CIK possibilities if multiple are listed.

        Args:
            soup (BeautifulSoup): BeautifulSoup object to search through.

        Returns:
            All possible companies that match lookup.
        """
        try:
            # Exclude table header
            table_rows = soup.find('table', {'summary': 'Results'}).find_all('tr')[1:]
            # Company names are in second column of table
            return [''.join(row.find_all('td')[1].find_all(string=True)) for row in table_rows]
        except AttributeError:
            # If there are no CIK possibilities, then no results were returned
            raise EDGARQueryError

    @staticmethod
    def _validate_cik(cik):
        """Check if CIK is 10 digit string."""
        if not (isinstance(cik, str) and len(cik) == 10 and cik.isdigit()):
            raise CIKError(cik)

    @staticmethod
    def _validate_lookup(lookup):
        """Ensure that lookup is string.

        Args:
            lookup: Value to lookup.

        Raises:
            TypeError: If lookup is not a non-empty string.
        """
        if not (lookup and isinstance(lookup, str)):
            raise TypeError("Lookup value must be string. Given type {0}.".format(type(lookup)))

    def get_ciks(self):
        """Validate lookup values and return corresponding CIKs.

        Returns:
            ciks (dict): Dictionary with lookup terms as keys and CIKs as values.

        """
        ciks = {}
        to_lookup = set(self.lookups)

        cik_map = get_cik_map(self.client.user_agent)

        # all keys upper case
        ticker_map = cik_map["ticker"]
        title_map = cik_map["title"]

        for lookup in to_lookup:
            lookup_norm = lookup.upper()
            if lookup_norm.isdigit():
                ciks[lookup] = lookup
            elif lookup_norm in ticker_map:
                ciks[lookup] = ticker_map[lookup_norm]
            elif lookup_norm in title_map:
                ciks[lookup] = title_map[lookup_norm]
            else:
                try:
                    result = self._get_cik_from_html(lookup)
                    self._validate_cik(result)  # raises CIKError if not valid CIK
                    ciks[lookup] = result
                except CIKError:
                    pass  # If multiple companies, found, print out warnings and skip
        return ciks



import datetime
from functools import reduce
from typing import Union

from secedgar.client import NetworkClient
from secedgar.core.daily import DailyFilings
from secedgar.core.quarterly import QuarterlyFilings
from secedgar.exceptions import EDGARQueryError, NoFilingsError
from secedgar.utils import add_quarter, get_month, get_quarter


def fill_days(start, end, include_start=False, include_end=False):
    """Get dates for days in between start and end date.

    Args:
        start (``datetime.date``): Start date.
        end (``datetime.date``): End date.
        include_start (bool, optional): Whether or not to include start date in range.
            Defaults to False.
        include_end (bool, optional): Whether or not to include end date in range.
            Defaults to False.

    Returns:
        list of ``datetime.date``: List of dates between ``start`` and ``end``.
    """
    start_range = 0 if include_start else 1
    end_range = (end - start).days + 1 if include_end else (end - start).days
    return [start + datetime.timedelta(days=d) for d in range(start_range, end_range)]


class ComboFilings:
    """Class for retrieving all filings between specified dates.

    Args:
        start_date (Union[str, ``datetime.datetime``, ``datetime.date``], optional): Date before
            which not to fetch reports. Stands for "date after."
            Defaults to None (will fetch all filings before ``end_date``).
        end_date (Union[str, ``datetime.datetime``, ``datetime.date``], optional):
            Date after which not to fetch reports.
            Stands for "date before." Defaults to today.
        user_agent (Union[str, NoneType]): Value used for HTTP header "User-Agent" for all requests.
            If given None, a valid client with user_agent must be given.
            See the SEC's statement on
            `fair access <https://www.sec.gov/os/accessing-edgar-data>`_
            for more information.
        client (Union[NoneType, secedgar.client.NetworkClient], optional): Client to use for
            fetching data. If None is given, a user_agent must be given to pass to
            :class:`secedgar.client.NetworkClient`.
            Defaults to ``secedgar.client.NetworkClient`` if none is given.
        entry_filter (function, optional): A boolean function to determine
            if the FilingEntry should be kept. Defaults to `lambda _: True`.
            The ``FilingEntry`` object exposes 7 variables which can be
            used to filter which filings to keep. These are "cik", "company_name",
            "form_type", "date_filed", "file_name", "path", and "num_previously_valid".
        balancing_point (int): Number of days from which to change lookup method from using
            ``DailyFilings`` to ``QuarterlyFilings``. If ``QuarterlyFilings`` is used, an
            additional filter will be added to limit which days are included.
            Defaults to 30.
        kwargs: Any keyword arguments to pass to ``NetworkClient`` if no client is specified.

    .. versionadded:: 0.4.0

    Examples:
        To download all filings from January 6, 2020 until November 5, 2020, you could do following:

        .. code-block:: python

            from datetime import date
            from secedgar import ComboFilings

            combo_filings = ComboFilings(start_date=date(2020, 1, 6),
                                         end_date=date(2020, 11, 5)
            combo_filings.save('/my_directory')
    """

    def __init__(self,
                 start_date: datetime.date,
                 end_date: datetime.date,
                 user_agent: Union[str, None] = None,
                 client=None,
                 entry_filter=lambda _: True,
                 balancing_point=30,
                 **kwargs):
        self.entry_filter = entry_filter
        self.start_date = start_date
        self.end_date = end_date
        self.user_agent = user_agent
        self._client = client or NetworkClient(user_agent=user_agent, **kwargs)
        self._balancing_point = balancing_point

    @property
    def entry_filter(self):
        """A boolean function to be tested on each listing entry.

        This is tested regardless of download method.
        """
        return self._entry_filter

    @entry_filter.setter
    def entry_filter(self, fn):
        if callable(fn):
            self._entry_filter = fn
        else:
            raise ValueError('entry_filter must be a function or lambda.')

    @property
    def client(self):
        """``secedgar.client.NetworkClient``: Client to use to make requests."""
        return self._client

    @property
    def start_date(self):
        """Union([datetime.date]): Date before which no filings fetched."""
        return self._start_date

    @start_date.setter
    def start_date(self, val):
        if val:
            self._start_date = val
        else:
            self._start_date = None

    @property
    def end_date(self):
        """Union([datetime.date]): Date after which no filings fetched."""
        return self._end_date

    @end_date.setter
    def end_date(self, val):
        self._end_date = val

    @property
    def balancing_point(self):
        """int: Point after which to use ``QuarterlyFilings`` with ``entry_filter`` to get data."""
        return self._balancing_point

    def _get_quarterly_daily_date_lists(self):
        """Break down date range into combination of quarters and single dates.

        Returns:
            tuple of lists: Quarterly date list and daily date list.
        """
        # Initialize quarter and date lists
        current_date = self.start_date
        quarterly_date_list = []
        daily_date_list = []

        while current_date <= self.end_date:
            current_quarter = get_quarter(current_date)
            current_year = current_date.year
            next_year, next_quarter = add_quarter(current_year, current_quarter)
            next_start_quarter_date = datetime.date(next_year, get_month(next_quarter), 1)

            days_till_next_quarter = (next_start_quarter_date -
                                      current_date).days
            days_till_end = (self.end_date - current_date).days

            # If there are more days until the end date than there are
            # in the quarter, add
            if days_till_next_quarter <= days_till_end:
                current_start_quarter_date = datetime.date(current_year,
                                                           get_month(current_quarter), 1)
                if current_start_quarter_date == current_date:
                    quarterly_date_list.append(
                        (current_year, current_quarter, lambda x: True))
                    current_date = next_start_quarter_date
                elif days_till_next_quarter > self.balancing_point:
                    quarterly_date_list.append(
                        (current_year, current_quarter,
                         lambda x: datetime.datetime.strptime(
                             x.date_filed, '%Y-%m-%d'
                         ).date() >= self.start_date))
                    current_date = next_start_quarter_date
                else:
                    daily_date_list.extend(fill_days(start=current_date,
                                                     end=next_start_quarter_date,
                                                     include_start=True,
                                                     include_end=False))
                    current_date = next_start_quarter_date
            else:
                if days_till_end > self.balancing_point:
                    if days_till_next_quarter - 1 == days_till_end:
                        quarterly_date_list.append(
                            (current_year, current_quarter, lambda x: True))
                        current_date = next_start_quarter_date
                    else:
                        quarterly_date_list.append(
                            (current_year, current_quarter,
                             lambda x: datetime.datetime.strptime(
                                 x.date_filed, '%Y-%m-%d'
                             ).date() <= self.end_date))
                        current_date = self.end_date
                else:
                    daily_date_list.extend(fill_days(start=current_date,
                                                     end=self.end_date,
                                                     include_start=True,
                                                     include_end=True))
                    break
        return quarterly_date_list, daily_date_list

    @property
    def quarterly_date_list(self):
        """List of tuples: List of tuples with year, quarter, and ``entry_filter`` to use."""
        return self._get_quarterly_daily_date_lists()[0]  # 0 = quarterly

    @property
    def daily_date_list(self):
        """List of ``datetime.date``: List of dates for which to fetch daily data."""
        return self._get_quarterly_daily_date_lists()[1]  # 1 = daily

    def get_urls(self):
        """Get all urls between ``start_date`` and ``end_date``."""
        # Use functools.reduce for speed
        # see https://stackoverflow.com/questions/10461531/merge-and-sum-of-two-dictionaries
        def _reducer(accumulator, dictionary):
            for key, value in dictionary.items():
                accumulator[key] = accumulator.get(key, []) + value
            return accumulator

        list_of_dicts = []
        for (year, quarter, f) in self.quarterly_date_list:
            q = QuarterlyFilings(year=year,
                                 quarter=quarter,
                                 user_agent=self.user_agent,
                                 client=self.client,
                                 entry_filter=lambda x: f(x) and self.entry_filter(x))
            list_of_dicts.append(q.get_urls())

        for _date in self.daily_date_list:
            d = DailyFilings(date=_date,
                             user_agent=self.user_agent,
                             client=self.client,
                             entry_filter=self.entry_filter)
            try:
                list_of_dicts.append(d.get_urls())
            except EDGARQueryError:  # continue if no URLs available for given day
                continue

        complete_dictionary = reduce(_reducer, list_of_dicts, {})
        return complete_dictionary

    def save(self,
             directory,
             dir_pattern=None,
             file_pattern="{accession_number}",
             download_all=False,
             daily_date_format="%Y%m%d"):
        """Save all filings between ``start_date`` and ``end_date``.

        Only filings that satisfy args given at initialization will
        be saved.

        Args:
            directory (str): Directory where filings should be stored.
            dir_pattern (str, optional): Format string for subdirectories. Defaults to None.
            file_pattern (str, optional): Format string for files. Defaults to "{accession_number}".
            download_all (bool, optional): Type of downloading system, if true downloads
                all data for each day, if false downloads each file in index.
                Defaults to False.
            daily_date_format (str, optional): Format string to use for the `{date}` pattern.
                Defaults to "%Y%m%d".
        """
        # Go through all quarters and dates and save filings using appropriate class
        for (year, quarter, f) in self.quarterly_date_list:
            q = QuarterlyFilings(year=year,
                                 quarter=quarter,
                                 user_agent=self.client.user_agent,
                                 client=self.client,
                                 entry_filter=lambda x: f(x) and self.entry_filter(x))
            q.save(directory=directory,
                   dir_pattern=dir_pattern,
                   file_pattern=file_pattern,
                   download_all=download_all)

        for date_ in self.daily_date_list:
            d = DailyFilings(date=date_,
                             user_agent=self.client.user_agent,
                             client=self.client,
                             entry_filter=self.entry_filter)
            try:
                d.save(directory=directory,
                       dir_pattern=dir_pattern,
                       file_pattern=file_pattern,
                       download_all=download_all,
                       date_format=daily_date_format)
            except (EDGARQueryError, NoFilingsError):  # continue if no filings for given day
                continue



import os
import platform
from datetime import date, datetime

import pytest

import secedgar.utils as utils


class TestUtils:
    @pytest.mark.parametrize(
        "bad_date",
        [
            "2012101",
            "201210",
            "1010",
            "2012011",
            2012011,
            2012101,
            2012,
            1010,
            201210
        ]
    )
    def test_bad_date_formats(self, bad_date):
        with pytest.raises(TypeError):
            utils.sanitize_date(bad_date)

    @pytest.mark.parametrize(
        "good_date",
        [
            "20120101",
            20120101
        ]
    )
    def test_good_formats_no_change(self, good_date):
        """Tests formats that should not change from what is given. """
        assert utils.sanitize_date(good_date) == good_date

    @pytest.mark.parametrize(
        "dt_date,expected",
        [
            (datetime(2018, 1, 1), "20180101"),
            (datetime(2020, 3, 4), "20200304"),
            (datetime(2020, 7, 18), "20200718"),
            (date(2018, 1, 1), "20180101"),
            (date(2020, 3, 4), "20200304"),
            (date(2020, 7, 18), "20200718")
        ]
    )
    def test_good_formats_datetime(self, dt_date, expected):
        assert utils.sanitize_date(dt_date) == expected

    @pytest.mark.parametrize(
        "date,expected_quarter",
        [
            (datetime(2020, 1, 1), 1),
            (datetime(2020, 2, 1), 1),
            (datetime(2020, 3, 1), 1),
            (datetime(2020, 3, 31), 1),
            (datetime(2020, 4, 1), 2),
            (datetime(2020, 5, 1), 2),
            (datetime(2020, 6, 1), 2),
            (datetime(2020, 6, 30), 2),
            (datetime(2020, 7, 1), 3),
            (datetime(2020, 8, 1), 3),
            (datetime(2020, 9, 1), 3),
            (datetime(2020, 9, 30), 3),
            (datetime(2020, 10, 1), 4),
            (datetime(2020, 11, 1), 4),
            (datetime(2020, 12, 1), 4),
            (datetime(2020, 12, 31), 4),
        ]
    )
    def test_get_quarter(self, date, expected_quarter):
        assert utils.get_quarter(date) == expected_quarter

    @pytest.mark.parametrize(
        "bad_quarter",
        [
            0.0,
            0,
            5.0,
            5,
            1.0,
            2.0
        ]
    )
    def test_get_month_bad_quarter(self, bad_quarter):
        with pytest.raises(TypeError):
            utils.get_month(bad_quarter)

    @pytest.mark.parametrize(
        "quarter,month",
        [
            (1, 1),
            (2, 4),
            (3, 7),
            (4, 10)
        ]
    )
    def test_get_month(self, quarter, month):
        assert utils.get_month(quarter) == month

    @pytest.mark.parametrize(
        "year,quarter,expected",
        [
            (2020, 1, (2020, 2)),
            (2020, 4, (2021, 1)),
            (2020, 2, (2020, 3)),
            (2020, 3, (2020, 4))
        ]
    )
    def test_add_quarter(self, year, quarter, expected):
        assert utils.add_quarter(year, quarter) == expected

    @pytest.mark.parametrize(
        "bad_quarter",
        [
            0,
            5
        ]
    )
    def test_add_quarter_bad_quarter(self, bad_quarter):
        with pytest.raises(TypeError):
            utils.add_quarter(2020, bad_quarter)

    @pytest.mark.skipif(platform.system() == "Windows",
                        reason="This test meant for Linux & Mac.")
    def test_make_path_expand_user(self):
        # make sure that you do not have a directory matching this if testing locally
        path_to_expand = "~/_____testing_____"
        utils.make_path(path_to_expand)
        path_expanded = os.path.expanduser(path_to_expand)
        try:
            assert os.path.exists(path_expanded)
        finally:
            os.rmdir(path_expanded)



import asyncio
import os
import warnings
from datetime import date

from secedgar.cik_lookup import CIKLookup
from secedgar.client import NetworkClient
from secedgar.core._base import AbstractFiling
from secedgar.core.filing_types import FilingType
from secedgar.exceptions import FilingTypeError
from secedgar.utils import sanitize_date


class CompanyFilings(AbstractFiling):
    """Base class for receiving EDGAR filings.

    Args:
        cik_lookup (str): Central Index Key (CIK) for company of interest.
        filing_type (Union[secedgar.core.filing_types.FilingType, None]): Valid filing type
            enum. Defaults to None. If None, then all filing types for CIKs will be returned.
        user_agent (Union[str, NoneType]): Value used for HTTP header "User-Agent" for all requests.
            If given None, a valid client with user_agent must be given.
            See the SEC's statement on
            `fair access <https://www.sec.gov/os/accessing-edgar-data>`_
            for more information.
        start_date (Union[str, datetime.datetime, datetime.date], optional): Date before
            which not to fetch reports. Stands for "date after."
            Defaults to None (will fetch all filings before ``end_date``).
        end_date (Union[str, datetime.datetime, datetime.date], optional):
            Date after which not to fetch reports.
            Stands for "date before." Defaults to today.
        count (int): Number of filings to fetch. Will fetch up to `count` if that many filings
            are available. Defaults to all filings available.
        ownership (str): Must be in {"include", "exclude"}. Whether or not to include ownership
            filings.
        match_format (str): Must be in {"EXACT", "AMEND", "ALL"}.
        kwargs: See kwargs accepted for :class:`secedgar.client.NetworkClient`.

    Examples:
        Restrict the start and end dates by using the ``start_date`` and ``end_date`` arguments.

        .. code-block:: python

            from secedgar import FilingType, CompanyFilings
            from datetime import date

            filing = CompanyFilings(cik_lookup="aapl",
                                    filing_type=FilingType.FILING_10K,
                                    start_date=date(2015, 1, 1),
                                    end_date=date(2019, 1, 1),
                                    user_agent="Name (email)")

        If you would like to find all filings from some ``start_date`` until today,
        simply exclude ``end_date``. The end date defaults to today's date.

        .. code-block:: python

            from secedgar import FilingType, CompanyFilings
            from datetime import date

            # end date defaults to today
            filing = CompanyFilings(cik_lookup="aapl",
                                    filing_type=FilingType.FILING_10K,
                                    start_date=date(2015, 1, 1),
                                    user_agent="Name (email)")

        You can also look up a specific filing type for multiple companies.

        .. code-block:: python

            from secedgar import FilingType, CompanyFilings
            from datetime import date

            filing = CompanyFilings(cik_lookup=["aapl", "msft", "fb"],
                                    filing_type=FilingType.FILING_10K,
                                    start_date=date(2015, 1, 1),
                                    user_agent="Name (email)")

        *For a full list of the available filing types, please see*
        :class:`secedgar.core.FilingType`.


        SEC requests that traffic identifies itself via a user agent string. You can
        customize this according to your preference using the ``user_agent`` argument.

        .. code-block:: python

            from secedgar import FilingType, CompanyFilings
            from datetime import date

            filing = CompanyFilings(cik_lookup=["aapl", "msft", "fb"],
                                    filing_type=FilingType.FILING_10K,
                                    start_date=date(2015, 1, 1),
                                    user_agent="Name (email)")

    .. versionadded:: 0.4.0
    """

    def __init__(self,
                 cik_lookup,
                 filing_type=None,
                 user_agent=None,
                 start_date=None,
                 end_date=date.today(),
                 client=None,
                 count=None,
                 ownership="include",
                 match_format="ALL",
                 **kwargs):
        self._params = {
            "action": "getcompany",
            "output": "xml",
            "start": 0
        }
        self.start_date = start_date
        self.end_date = end_date
        self.filing_type = filing_type
        self.count = count
        self.match_format = match_format
        # Make default client NetworkClient and pass in kwargs
        self._client = client or NetworkClient(user_agent=user_agent, **kwargs)
        # make CIKLookup object for users if not given
        self.cik_lookup = cik_lookup
        self.ownership = ownership

    @property
    def path(self):
        """str: Path added to client base."""
        return "cgi-bin/browse-edgar"

    @property
    def params(self):
        """:obj:`dict`: Parameters to include in requests."""
        if self.start_date:
            self._params["datea"] = sanitize_date(self.start_date)
        else:
            self._params.pop("datea", None)  # if no start date, make sure it isn't in params

        if self.end_date:
            self._params["dateb"] = sanitize_date(self.end_date)
        else:
            self._params.pop("dateb", None)  # if no end date, make sure it isn't in params

        self._params["ownership"] = self.ownership
        return self._params

    @property
    def client(self):
        """``secedgar.client.NetworkClient``: Client to use to make requests."""
        return self._client

    @property
    def start_date(self):
        """Union([datetime.date, datetime.datetime, str]): Date before which no filings fetched."""
        return self._start_date

    @property
    def match_format(self):
        """The match format to use when searching for filings."""
        return self._match_format

    @match_format.setter
    def match_format(self, val):
        if val not in ("EXACT", "AMEND", "ALL"):
            raise ValueError("Format must be 'EXACT', 'AMEND', or 'ALL'.")
        self._match_format = val

    @start_date.setter
    def start_date(self, val):
        sanitize_date(val)  # make sure start date is valid
        self._start_date = val

    @property
    def end_date(self):
        """Union([datetime.date, datetime.datetime, str]): Date after which no filings fetched."""
        return self._end_date

    @end_date.setter
    def end_date(self, val):
        sanitize_date(val)  # make sure end date is valid
        self._end_date = val

    @property
    def filing_type(self):
        """``secedgar.core.FilingType``: FilingType enum of filing."""
        return self._filing_type

    @filing_type.setter
    def filing_type(self, filing_type):
        if isinstance(filing_type, FilingType):
            self._params["type"] = filing_type.value
        elif filing_type is not None:
            raise FilingTypeError(
                f"filing_type must be of type FilingType. Given {type(filing_type)}."
            )
        self._filing_type = filing_type

    @property
    def count(self):
        """Number of filings to fetch."""
        return self._count

    @count.setter
    def count(self, val):
        if val is None:
            self._count = None
        elif not isinstance(val, int):
            raise TypeError("Count must be positive integer or None.")
        elif val < 1:
            raise ValueError("Count must be positive integer or None.")
        else:
            self._count = val
            self._params["count"] = val

    @property
    def cik_lookup(self):
        """``secedgar.cik_lookup.CIKLookup``: CIKLookup object."""
        return self._cik_lookup

    @cik_lookup.setter
    def cik_lookup(self, val):
        if not isinstance(val, CIKLookup):
            val = CIKLookup(val, client=self.client)
        self._cik_lookup = val

    @property
    def ownership(self):
        """str: Whether or not to include ownership in search."""
        return self._ownership

    @ownership.setter
    def ownership(self, val):
        if val not in ("include", "exclude"):
            raise ValueError(f"Ownership must be 'include' or 'exclude'. Given {val}.")
        self._ownership = val

    def get_urls(self, **kwargs):
        """Get urls for all CIKs given to Filing object.

        Args:
            **kwargs: Anything to be passed to requests when making get request.
                See keyword arguments accepted for
                ``secedgar.client.NetworkClient.get_soup``.

        Returns:
            urls (list): List of urls for txt files to download.
        """
        return {
            key: self._get_urls_for_cik(cik, **kwargs)
            for key, cik in self.cik_lookup.lookup_dict.items()
        }

    def _filter_filing_links(self, data):
        """Filter filing links from data to only include exact matches.

        Args:
            data (``bs4.BeautifulSoup``): ``BeautifulSoup`` object to parse.

        Returns:
            list of str: List of URLs that match ``filing_type``.
        """
        filings = data.find_all("filing")
        if self.filing_type is None:
            return [f.filinghref.string for f in filings]
        else:
            return [f.filinghref.string for f in filings if f.type.string == self.filing_type.value]

    # TODO: Change this to return accession numbers that are turned into URLs later
    def _get_urls_for_cik(self, cik, **kwargs):
        """Get all urls for specific company according to CIK.

        Must match start date, end date, filing_type, and count parameters.

        Args:
            cik (str): CIK for company.
            **kwargs: Anything to be passed to requests when making get request.
                See keyword arguments accepted for
                ``secedgar.client.NetworkClient.get_soup``.

        Returns:
            txt_urls (list of str): Up to the desired number of URLs for that specific company
            if available.
        """
        self.params["CIK"] = cik
        links = []
        self.params["start"] = 0  # set start back to 0 before paginating
        while self.count is None or len(links) < self.count:
            data = self.client.get_soup(self.path, self.params, **kwargs)
            filtered_links = self._filter_filing_links(data)
            links.extend(filtered_links)
            self.params["start"] += self.client.batch_size
            if len(data.find_all("filinghref")) == 0:  # no more filings
                break

        txt_urls = [link[:link.rfind("-")].strip() + ".txt" for link in links]

        if isinstance(self.count, int) and len(txt_urls) < self.count:
            warnings.warn(
                "Only {num} of {count} filings were found for {cik}.".format(
                    num=len(txt_urls), count=self.count, cik=cik))
        # Takes `count` filings at most
        return txt_urls[:self.count]

    def save(self, directory, dir_pattern=None, file_pattern=None):
        """Save files in specified directory.

        Each txt url looks something like:
        https://www.sec.gov/Archives/edgar/data/1018724/000101872419000043/0001018724-19-000043.txt

        Args:
            directory (str): Path to directory where files should be saved.
            dir_pattern (str): Format string for subdirectories. Default is "{cik}/{type}".
                Valid options are {cik} and/or {type}.
            file_pattern (str): Format string for files. Default is "{accession_number}".
                Valid options are {accession_number}.

        Returns:
            None

        Raises:
            ValueError: If no text urls are available for given filing object.
        """
        urls = self.get_urls_safely()

        if dir_pattern is None:
            dir_pattern = os.path.join("{cik}", "{type}")
        if file_pattern is None:
            file_pattern = "{accession_number}"

        inputs = []
        for cik, links in urls.items():
            # If no filing type, just leave all filings for CIK under directory named after CIK
            if self.filing_type is None:
                formatted_dir = cik
            else:
                formatted_dir = dir_pattern.format(cik=cik,
                                                   type=self.filing_type.value)
            for link in links:
                formatted_file = file_pattern.format(
                    accession_number=self.get_accession_number(link))
                path = os.path.join(directory, formatted_dir, formatted_file)
                inputs.append((link, path))

        asyncio.run(self.client.wait_for_download_async(inputs))



import os

import requests


def datapath(*args):
    """Get the path to a data file.

    Returns:
        path including ``secedgar/tests/data``.
    """
    base_path = os.path.join(os.path.dirname(__file__), 'data')
    return os.path.join(base_path, *args)


class MockResponse(requests.Response):
    def __init__(self, datapath_args=[],
                 status_code=200,
                 content=None,
                 *args,
                 **kwargs):
        super().__init__()
        self.status_code = status_code
        if content is not None:
            self._content = content
        else:
            with open(datapath(*datapath_args), "rb", **kwargs) as f:
                self._content = f.read()
        self._content_consumed = True

    def __call__(self, *args, **kwargs):
        return self


class AsyncMockResponse(MockResponse):
    def __init__(self, datapath_args=[],
                 status_code=200,
                 content=None,
                 encoding="utf-8",
                 *args,
                 **kwargs):
        super().__init__(datapath_args=datapath_args,
                         status_code=status_code,
                         content=content,
                         *args,
                         **kwargs)
        self._encoding = encoding

    async def __aexit__(self, exc_type, exc, tb):
        pass

    async def __aenter__(self):
        return self

    async def read(self):
        return self._content


class AsyncLimitedResponsesSession:
    def __init__(self, response=AsyncMockResponse(content=bytes("Testing...", "utf-8")), limit=10):
        self._response = response
        self._limit = limit
        self._requests_made = 0

    async def get(self, *args, **kwargs):
        if self._requests_made < self._limit:
            return await self._response
        return

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc, tb):
        pass



import datetime
import os
from datetime import date

import pytest
from click.testing import CliRunner

from secedgar.cli import cli, date_cleanup
from secedgar.exceptions import FilingTypeError
# Borrow mocks without including in conftest
from secedgar.tests.core.test_daily import mock_daily_idx_file  # noqa: F401
from secedgar.tests.core.test_daily import \
    mock_daily_quarter_directory  # noqa: F401


def run_cli_command(cli_instance,
                    user_input,
                    directory=None,
                    user_agent="'My User Agent (email@example.com)'",
                    catch_exceptions=False):
    runner = CliRunner()
    user_input = user_input.split()
    user_input = ['--user-agent', user_agent] + user_input + ['--directory', directory]
    return runner.invoke(cli_instance, user_input, catch_exceptions=catch_exceptions)


def check_bad_inputs(cli_instance,
                     user_input,
                     expected_exception,
                     directory,
                     user_agent="'My User Agent (email@example.com)'"):
    # SystemExit does not raise exception by runner
    if expected_exception is SystemExit:
        result = run_cli_command(cli_instance=cli_instance,
                                 user_input=user_input,
                                 directory=directory,
                                 user_agent=user_agent)
        assert result.exit_code != 0
    else:
        with pytest.raises(expected_exception):
            run_cli_command(cli_instance=cli_instance,
                            user_input=user_input,
                            directory=directory,
                            user_agent=user_agent)


class TestCLI:
    @pytest.mark.parametrize(
        "user_input,expected_exception",
        [
            ("filing -l aapl msft Facebook", SystemExit),  # missing filing type
            ("filing -l aapl -t null", FilingTypeError),  # unrecognized filing type
            ("filing -l aapl -t FILING_10Q -n abc", SystemExit),  # count is not int
            ("filing -l aapl -t FILING_10Q -n 0",
             ValueError),  # no filings available if 0 picked
        ]
    )
    def test_cli_filing_bad_inputs(self, user_input, expected_exception, tmp_data_directory):
        check_bad_inputs(cli,
                         user_input=user_input,
                         expected_exception=expected_exception,
                         directory=tmp_data_directory)

    def test_cli_filing_multiple_companies_input(
            self,
            tmp_data_directory,
            mock_cik_validator_get_multiple_ciks,
            mock_single_cik_filing,
            mock_filing_response):
        result = run_cli_command(
            cli, "filing -l aapl -l msft -l amzn -t FILING_10Q -n 10", tmp_data_directory)
        assert result.exit_code == 0
        txt_files = [f for *_, files in os.walk(tmp_data_directory) for f in files]
        assert len(txt_files) == 30

    @pytest.mark.smoke
    @pytest.mark.parametrize(
        "user_input,count",
        [
            ("filing -l aapl -l msft -l amzn -t FILING_10Q -n 10", 10),
        ]
    )
    def test_cli_filing_multiple_companies_input_smoke(
            self,
            user_input,
            count,
            tmp_data_directory):

        result = run_cli_command(cli, user_input, tmp_data_directory)
        assert result.exit_code == 0
        txt_files = [f for *_, files in os.walk(tmp_data_directory) for f in files]
        if count is None:
            assert len(txt_files) == 3
        else:
            assert len(txt_files) == 3 * count

    @pytest.mark.parametrize(
        "user_input,expected_exception",
        [
            ("daily", SystemExit),
            ("daily -d 2020", ValueError)
        ]
    )
    def test_cli_daily_bad_inputs(self, user_input, expected_exception, tmp_data_directory):
        check_bad_inputs(cli,
                         user_input=user_input,
                         expected_exception=expected_exception,
                         directory=tmp_data_directory)

    @pytest.mark.parametrize(
        "user_input",
        [
            "daily -d 20201113",
            "filing -l aapl -l msft -l amzn -t FILING_10Q -n 10"
        ]
    )
    def test_cli_requires_user_agent(self, user_input, tmp_data_directory):
        check_bad_inputs(cli, user_input, SystemExit, tmp_data_directory, user_agent=None)

    @pytest.mark.parametrize(
        "input,output",
        [
            ("20210423", date(2021, 4, 23)),
            ("19900101", date(1990, 1, 1)),
            ("20000101", date(2000, 1, 1)),
            (None, None)
        ]

    )
    def test_date_cleanup(self, input, output):
        result = date_cleanup(input)
        if isinstance(input, str):
            assert isinstance(result, datetime.date)
        else:
            assert result is None
        assert result == output



from datetime import date

import pytest
from secedgar.client import NetworkClient
from secedgar.core.combo import ComboFilings, fill_days


def lambda_matches(a, b):
    return a.__code__.co_code == b.__code__.co_code


def line_matches(a, b):
    return a[0] == b[0] and a[1] == b[1] and lambda_matches(a[2], b[2])


def quarterly_list_matches(l1, l2):
    return all(line_matches(a, b) for (a, b) in zip(l1, l2))


class TestComboFilings:
    @pytest.mark.parametrize(
        "include_start,include_end,expected",
        [
            (True, True, [date(2020, 1, i) for i in (1, 2, 3)]),
            (True, False, [date(2020, 1, i) for i in (1, 2)]),
            (False, False, [date(2020, 1, 2)]),
            (False, True, [date(2020, 1, i) for i in (2, 3)]),
        ]
    )
    def test_fill_days(self, include_start, include_end, expected):
        result = fill_days(start=date(2020, 1, 1),
                           end=date(2020, 1, 3),
                           include_start=include_start,
                           include_end=include_end)
        assert result == expected

    def test_user_agent_client_none(self):
        with pytest.raises(TypeError):
            _ = ComboFilings(start_date=date(2020, 1, 1),
                             end_date=date(2020, 12, 31),
                             user_agent=None,
                             client=None)

    def test_client_passed_to_objects(self, mock_user_agent):
        client = NetworkClient(user_agent=mock_user_agent)
        combo = ComboFilings(start_date=date(2020, 1, 1),
                             end_date=date(2020, 12, 31),
                             client=client)
        assert combo.client == client

    def test_combo_quarterly_only_one_year(self, mock_user_agent):
        combo = ComboFilings(start_date=date(2020, 1, 1),
                             end_date=date(2020, 12, 31),
                             user_agent=mock_user_agent)
        expected = [(2020, 1, lambda _: True), (2020, 2, lambda _: True),
                    (2020, 3, lambda _: True), (2020, 4, lambda _: True)]
        assert quarterly_list_matches(combo.quarterly_date_list, expected)
        assert len(combo.daily_date_list) == 0

    def test_combo_quarterly_only_multiple_years(self, mock_user_agent):
        combo = ComboFilings(start_date=date(2018, 10, 1),
                             end_date=date(2020, 6, 30),
                             user_agent=mock_user_agent)
        expected = [
            (2018, 4, lambda _: True),
            (2019, 1, lambda _: True),
            (2019, 2, lambda _: True),
            (2019, 3, lambda _: True),
            (2019, 4, lambda _: True),
            (2020, 1, lambda _: True),
            (2020, 2, lambda _: True),
        ]
        assert quarterly_list_matches(combo.quarterly_date_list, expected)
        assert len(combo.daily_date_list) == 0

    def test_combo_daily_only_single_day(self, mock_user_agent):
        combo = ComboFilings(start_date=date(2020, 12, 10),
                             end_date=date(2020, 12, 10),
                             user_agent=mock_user_agent)
        assert [str(s) for s in combo.daily_date_list] == ["2020-12-10"]
        assert len(combo.quarterly_date_list) == 0

    def test_combo_daily_only_multiple_days(self, mock_user_agent):
        combo = ComboFilings(start_date=date(2020, 12, 10),
                             end_date=date(2020, 12, 12),
                             user_agent=mock_user_agent)
        expected = ["2020-12-10", "2020-12-11", "2020-12-12"]
        assert [str(s) for s in combo.daily_date_list] == expected
        assert len(combo.quarterly_date_list) == 0

    @pytest.mark.parametrize(
        "start_date,end_date,quarterly_expected,daily_expected", [
            (date(2019, 12, 28), date(
                2020, 4, 1), [(2020, 1, lambda _: True)], [
                    "2019-12-28", "2019-12-29", "2019-12-30", "2019-12-31",
                    "2020-04-01"
            ]),
            (date(2020, 3, 30), date(2020, 10, 2), [
                (2020, 2, lambda _: True), (2020, 3, lambda _: True)
            ], ["2020-03-30", "2020-03-31", "2020-10-01", "2020-10-02"]),
            (date(2020, 1, 1), date(2020, 4, 2), [
                (2020, 1, lambda _: True)
            ], ["2020-04-01", "2020-04-02"]),
            (date(2020, 3, 30), date(2020, 9, 30), [
                (2020, 2, lambda _: True), (2020, 3, lambda _: True)
            ], ["2020-03-30", "2020-03-31"]),
            # (date(2020, 1, 1), date(2020, 6, 28), [], [(2020, 1, lambda _: True),
            #                                            (2020, 2, lambda x: date(x['date_filed']) <= date(2020, 6, 28))])  # noqa
        ])
    def test_combo_daily_quarterly_mixed(self, start_date,
                                         end_date,

                                         quarterly_expected,
                                         daily_expected,
                                         mock_user_agent):
        combo = ComboFilings(start_date=start_date, end_date=end_date, user_agent=mock_user_agent)
        assert [str(s) for s in combo.daily_date_list] == daily_expected
        assert quarterly_list_matches(combo.quarterly_date_list, quarterly_expected)

    def test_properties_on_init(self, mock_user_agent):
        start = date(2020, 1, 1)
        end = date(2020, 5, 30)
        bp = 25
        combo = ComboFilings(start_date=start,
                             end_date=end,
                             user_agent=mock_user_agent,
                             balancing_point=bp)

        assert combo.start_date == start
        assert combo.end_date == end
        assert combo.balancing_point == bp
        assert combo.client.user_agent == mock_user_agent

    @pytest.mark.parametrize(
        "bad_entry_filter",
        [
            None,
            [],
            (),
            0,
            ""]
    )
    def test_bad_entry_filter(self, bad_entry_filter):
        with pytest.raises(ValueError):
            _ = ComboFilings(start_date=date(2020, 1, 1),
                             end_date=date(2020, 5, 30),
                             entry_filter=bad_entry_filter)

    @pytest.mark.parametrize(
        "good_entry_filter",
        [
            lambda x: True,
            lambda x: False,
            lambda f: f.form_type.lower() == "10-k"
        ]
    )
    def test_good_entry_filter(self, good_entry_filter, mock_user_agent):
        combo = ComboFilings(date(2020, 1, 1),
                             date(2020, 5, 30),
                             entry_filter=good_entry_filter,
                             user_agent=mock_user_agent)
        assert combo.entry_filter == good_entry_filter

    def test_client_read_only(self, mock_user_agent):
        combo = ComboFilings(start_date=date(2020, 1, 1),
                             end_date=date(2020, 1, 3),
                             user_agent=mock_user_agent)
        with pytest.raises(AttributeError):
            combo.client = None

    def test_balancing_point_read_only(self, mock_user_agent):
        combo = ComboFilings(start_date=date(2020, 1, 1),
                             end_date=date(2020, 1, 3),
                             user_agent=mock_user_agent)
        with pytest.raises(AttributeError):
            combo.balancing_point = 20

    def test_combo_filings_quarterly_lambda(self, mock_user_agent):
        # Tests issue raised in GH#269.
        combo = ComboFilings(start_date=date(2022, 2, 28),
                             end_date=date(2022, 4, 15),
                             user_agent=mock_user_agent)

        # Make sure you can run ``_get_quarterly_daily_date_lists``
        quarter_list, date_list = combo._get_quarterly_daily_date_lists()

        assert len(quarter_list) > 0, "Expected quarterly list not to be empty"
        assert len(date_list) > 0, "Expected date list not to be empty"
        assert date_list == [date(2022, 4, d) for d in range(1, 16)], """
        Date list should have April 1 through 15."""
        assert len(quarter_list) == 1, "Should only have one quarter"
        assert quarter_list[0][0] == 2022, "Only quarter in list should be Q1 2022"
        assert quarter_list[0][1] == 1, "Only quarter in list should be Q1 2022"

    def test_combo_no_user_agent_raises_error(self):
        with pytest.raises(TypeError):
            _ = ComboFilings(start_date=date(2022, 2, 28),
                             end_date=date(2022, 4, 15))



from typing import List, Union

import requests

from secedgar.cik_lookup import CIKLookup

API_BASE = "https://data.sec.gov/api/"
XBRL_BASE = "{0}xbrl/".format(API_BASE)


def _get_lookup_dict(lookups: List[str], user_agent: str):
    """Utility function to get CIKs for lookups.

    Args:
        lookups (list of str): List of tickers or company names to get CIKs for.
        user_agent (str): User agent to give to SEC.

    Returns:
        dict: Dictionary with lookup and CIK key-value pairs.
    """
    cik_lookup = CIKLookup(lookups=lookups,
                           user_agent=user_agent,
                           client=None)
    return cik_lookup.lookup_dict


def _combine_dicts(*dicts):
    """Utility function to combine dictionary values when values are lists.

    Returns:
        dict: Dictionaries combined into one.

    Examples:
        >>> a = {"A": [1, 2, 3], "B": [4, 5, 6]}
        >>> b = {"A": [7, 8], "B": [0, 1, 2]}
        >>> _combine_dicts(a, b)
        {'A': [1, 2, 3, 7, 8], 'B': [4, 5, 6, 0, 1, 2]}
        >>> _combine_dicts(a)
        {'A': [1, 2, 3], 'B': [4, 5, 6]}
    """
    final = {}
    for d in dicts:
        for k, v in d.items():
            if k in final:
                final[k] += v
            else:
                final[k] = v
    return final


def get_submissions(lookups: Union[List[str], str],
                    user_agent: str,
                    recent: bool = True) -> dict:
    """Get information about submissions for entities.

    Args:
        lookups (list of str or str): Tickers or CIKs to get submission data for.
        user_agent (str): User agent to provide the SEC.
        recent (bool, optional): Whether or not to only get the most recent.
            Setting ``recent`` to True will give at least one year of filings or 1000 filings
            (whichever is more). Setting ``recent`` to False will return all filings.
            Defaults to True.

    Returns:
        dict: Dictionary with keys being the lookups and values being the responses from the API.
    """
    lookup_dict = _get_lookup_dict(lookups=lookups, user_agent=user_agent)
    submissions_dict = dict()
    url_base = "https://data.sec.gov/submissions/"
    for lookup, cik in lookup_dict.items():
        resp = requests.get("{0}CIK{1}.json".format(url_base, cik.zfill(10)),
                            headers={"user-agent": user_agent})
        resp_json = resp.json()
        if not recent:
            try:
                older_submission_files = resp_json["filings"]["files"]
            except KeyError:
                pass

            # Get data for older submission files and add to recent
            older_submissions = [requests.get("{0}{1}".format(url_base, f["name"]),
                                              headers={"user-agent": user_agent}).json()
                                 for f in older_submission_files]

            resp_json["filings"]["recent"] = _combine_dicts(resp_json["filings"]["recent"],
                                                            *older_submissions)
        submissions_dict[lookup] = resp_json
    return submissions_dict


def get_company_concepts(lookups: Union[List[str], str],
                         user_agent: str,
                         concept_name: str) -> dict:
    """Get company concepts using SEC's REST API.

    Args:
        lookups (list of str or str): Tickers or CIKs to get concepts for.
        user_agent (str): User agent to send to SEC.
        concept_name (str): Name of the concept to get data for.

    Returns:
        dict: Dictionary with concept data for given lookups.

    Example:
        .. code::

            concept = "AccountsPayableCurrent"
            get_company_concepts(lookups=["AAPL"],
                                 user_agent="Name (email@example.com)",
                                 concept_name=concept)
    """
    lookup_dict = _get_lookup_dict(lookups=lookups, user_agent=user_agent)
    company_concepts = dict()
    for lookup, cik in lookup_dict.items():
        url = "{0}companyconcept/CIK{1}/us-gaap/{2}.json".format(
            XBRL_BASE,
            cik.zfill(10),
            concept_name
        )
        resp = requests.get(url,
                            headers={"user-agent": user_agent})
        company_concepts[lookup] = resp.json()
    return company_concepts


def get_company_facts(lookups: Union[List[str], str], user_agent: str) -> dict:
    """Get company facts for lookups.

    Args:
        lookups (list of str or str): Tickers or CIKs to get company facts for.
        user_agent (str): User agent to send to SEC.

    Returns:
        dict: Dictionary with lookups as keys and company fact dictionaries as values.

    Examples:
        >>> facts = get_company_facts(["aapl"], "Name (email@example.com)")
        >>> single_fact = facts["aapl"]["facts"]["us-gaap"]["Assets"]["units"]["USD"][0]
        >>> single_fact["val"]
        39572000000
        >>> single_fact["fy"]
        2009
        >>> single_fact["fp"]
        Q3
        >>> single_fact["form"]
        '10-Q'
        >>> single_fact["filed"]
        '2009-07-22'
    """
    lookup_dict = _get_lookup_dict(lookups=lookups, user_agent=user_agent)
    company_facts = dict()
    for lookup, cik in lookup_dict.items():
        url = "{0}companyfacts/CIK{1}.json".format(XBRL_BASE, cik.zfill(10))
        resp = requests.get(url, headers={"user-agent": user_agent})
        company_facts[lookup] = resp.json()
    return company_facts


def get_xbrl_frames(user_agent: str,
                    concept_name: str,
                    year: int,
                    quarter: Union[None, int] = None,
                    currency: str = "USD",
                    instantaneous: bool = False) -> dict:
    """Get data for concept name in year (and quarter, if given).

    Args:
        user_agent (str): User agent to send to the SEC.
        concept_name (str): Concept name to get. For example, "Assets".
        year (int): Year to get concept data for.
        quarter (Union[int, NoneType], optional): Quarter to get data for. If given None,
            will look for data for the entire year. Defaults to None.
        instantaneous (bool, optional): Whether to look for instantaneous data.
            See `SEC website for more <https://www.sec.gov/edgar/sec-api-documentation>`_.
            Defaults to False.

    Returns:
        dict: Dictionary with information about concept_name.
            Dig into the data key for all period data.

    Examples:
        .. code::

            frames = get_xbrl_frames(user_agent="Name (email@example.com)",
                                     concept_name="Assets",
                                     year=2020,
                                     quarter=3,
                                     instantaneous=True)
            print(frames["data"][-1]["entityName"])
            # Prints "MOXIAN (BVI) INC"
            print(frames["data"][-1]["val"])
            # Prints 2295657
            print(frames["data"][-1]["accn"])
            # Prints "0001493152-22-013323"
    """
    # Create URL
    period = "CY{0}".format(year) if quarter is None else "CY{0}Q{1}".format(year, quarter)
    if instantaneous:
        period += "I"
    url = "{0}frames/us-gaap/{1}/{2}/{3}.json".format(XBRL_BASE, concept_name, currency, period)
    # Request and add to dictionary
    resp = requests.get(url, headers={"user-agent": user_agent})
    xbrl_frames = resp.json()
    return xbrl_frames



import os
from datetime import date

from secedgar.core._index import IndexFilings
from secedgar.utils import get_quarter


class QuarterlyFilings(IndexFilings):
    """Class for retrieving all filings from specific year and quarter.

    Args:
        year (int): Must be in between 1993 and the current year (inclusive).
        quarter (int): Must be 1, 2, 3, or 4. Quarter of filings to fetch.
        user_agent (Union[str, NoneType], optional): Value used for HTTP header
            "User-Agent" for all requests. If given None, a valid client with
            user_agent must be given. See the SEC's statement on
            `fair access <https://www.sec.gov/os/accessing-edgar-data>`_
            for more information. Defaults to None.
        client (Union[NoneType, secedgar.client.NetworkClient], optional): Client to use for
            fetching data. If None is given, a user_agent must be given to pass to
            :class:`secedgar.client.NetworkClient`. Defaults to ``secedgar.client.NetworkClient``
            if none is given.
        entry_filter (function, optional): A boolean function to determine
            if the FilingEntry should be kept. Defaults to ``lambda _: True``.
            See :class:`secedgar.core.DailyFilings` for more detail.
        kwargs: Keyword arguments to pass to ``secedgar.core._index.IndexFilings``.

    Examples:
        .. code-block:: python

            import secedgar as sec
            quarterly = sec.QuarterlyFilings(year=2021,
                                             quarter=2,
                                             user_agent="Name (email)")
            quarterly.save("/path/to/dir")

    """

    def __init__(self,
                 year,
                 quarter,
                 user_agent=None,
                 client=None,
                 entry_filter=lambda _: True,
                 **kwargs):
        super().__init__(user_agent=user_agent,
                         client=client,
                         entry_filter=entry_filter,
                         **kwargs)
        self.year = year
        self.quarter = quarter

    @property
    def path(self):
        """Path property to pass to client."""
        return "Archives/edgar/full-index/{year}/QTR{num}/".format(
            year=self._year, num=self._quarter)

    @property
    def year(self):
        """Year of filings."""
        return self._year

    @year.setter
    def year(self, val):
        if not isinstance(val, int):
            raise TypeError("Year must be an integer.")
        elif val < 1993 or val > date.today().year:
            raise ValueError(
                "Year must be in between 1993 and {now} (inclusive)".format(
                    now=date.today().year))
        self._year = val

    @property
    def quarter(self):
        """Quarter of filings."""
        return self._quarter

    @quarter.setter
    def quarter(self, val):
        if not isinstance(val, int):
            raise TypeError("Quarter must be integer.")
        elif val not in range(1, 5):
            raise ValueError("Quarter must be in between 1 and 4 (inclusive).")
        elif self.year == date.today().year and val > get_quarter(date.today()):
            raise ValueError("Latest quarter for current year is {qtr}".format(
                qtr=get_quarter(date.today())))
        self._quarter = val

    @property
    def idx_filename(self):
        """Main index filename to look for."""
        return "master.idx"

    def _get_tar_urls(self):
        """The list of .tar.gz daily files in the current quarter."""
        soup = self.client.get_soup(self.tar_path, {})
        files = [a.get('href') for a in soup.find_all('a') if "nc.tar.gz" in a.get('href')]
        return files

    def save(self,
             directory,
             dir_pattern=None,
             file_pattern="{accession_number}",
             download_all=False):
        """Save all daily filings.

        Creates subdirectory within given directory of the form <YEAR>/QTR<QTR NUMBER>/.
        Then each distinct company name receives its own directory with all of its filings.
        See ``secedgar.core._index.IndexFilings`` for more detail.

        Args:
            directory (str): Directory where filings should be stored. Will be broken down
                further by company name and form type.
            dir_pattern (str): Format string for subdirectories. Default is
                `{year}/QTR{quarter}/{cik}`. Valid options to mix and match
                 are `{year}`, `{quarter}`, and `{cik}`.
            file_pattern (str): Format string for files. Default is `{accession_number}`.
                Valid options are `{accession_number}`.
            download_all (bool): Type of downloading system, if true downloads all data for each
                day, if false downloads each file in index. Default is `False`.
        """
        if dir_pattern is None:
            # https://stackoverflow.com/questions/11283961/partial-string-formatting
            dir_pattern = os.path.join('{year}', 'QTR{quarter}', '{cik}')

        # If "{cik}" is in dir_pattern, it will be passed on and if not it will be ignored
        formatted_dir = dir_pattern.format(year=self.year,
                                           quarter=self.quarter,
                                           cik="{cik}")
        self._save_filings(directory,
                           dir_pattern=formatted_dir,
                           file_pattern=file_pattern,
                           download_all=download_all)



from datetime import date, timedelta

from secedgar.core.combo import ComboFilings
from secedgar.core.company import CompanyFilings
from secedgar.core.daily import DailyFilings
from secedgar.core.filing_types import FilingType
from secedgar.core.quarterly import QuarterlyFilings
from secedgar.exceptions import FilingTypeError
from secedgar.utils import add_quarter, get_month, get_quarter


def filings(
    cik_lookup=None,
    filing_type=None,
    user_agent=None,
    start_date=None,
    end_date=date.today(),
    count=None,
    client=None,
    entry_filter=lambda _: True,
    **kwargs
):
    """Utility method to get best filing object.

    Args:
        cik_lookup (str): Central Index Key (CIK) for company of interest.
        start_date (datetime.date, optional): Date of daily filing to fetch.
        end_date (datetime.date, optional): Date of daily filing to fetch.
        filing_type (secedgar.core.filing_types.FilingType, optional): Valid filing type
            enum. Defaults to None. If None, then all filing types for CIKs will be returned.
        count (int, optional): Number of filings to fetch. Will fetch up to `count` if that
        many filings are available. Defaults to all filings available.
        client (secedgar.client.NetworkClient, optional): Client to use. Defaults to
                    ``secedgar.client.NetworkClient`` if None given.
        entry_filter (function, optional): A boolean function to determine
            if the FilingEntry should be kept. Defaults to ``lambda _: True``.
            See :class:`secedgar.core.DailyFilings` for more detail.
        kwargs: Any keyword arguments to pass to ``NetworkClient`` if no client is specified.

    Examples:
        Using the ``filings`` function from secedgar is the easiest way to retrieve filings.

        Depending on the arguments given, secedgar will return an object that will get you
        the information you want from EDGAR.

        There are 4 main classes which can be returned.

            - :class:`secedgar.ComboFilings` for fetching filings over multiple days
              that does not fall exactly into a quarter
            - :class:`secedgar.CompanyFilings` for fetching a particular
              filing type for one or more companies
            - :class:`secedgar.DailyFilings` for fetching all filings
              from a specific date
            - :class:`secedgar.QuarterlyFilings` for fetching all filings
              from a specific quarter

        To get all filings over a time span, you could use something like below.

        .. code-block:: python

            from datetime import date
            from secedgar import filings, FilingType

            # secedgar creates correct filing object for given arguments
            # this will fetch the first 50 filings found over the time span
            my_filings = filings(start_date=date(2020, 12, 10),
                                 end_date=date(2020, 12, 15),
                                 filing_type=FilingType.FILING_4,
                                 user_agent="Name (email)",
                                 count=50)

            # easy access to methods shared across all 4 different filing classes
            my_filings_urls = my_filings.get_urls()
            my_filings.save("/path/to/directory")

        To get a single filing type for one or more companies, you could use this:

        .. code-block:: python

            from secedgar import filings, FilingType

            # similar to above, but fetches filings for specific tickers
            company_filings = filings(cik_lookup=["aapl", "fb"],
                                      filing_type=sec.FilingType.FILING_10Q,
                                      user_agent="Name (email)")
            company_filings_urls = company_filings.get_urls()
            company_filings.save("/path/to/directory")

        To get filings for a single day, you could use something like this:

        .. code-block:: python

            from datetime import date
            from secedgar import filings

            # all filings for
            daily_filings = filings(start_date=date(2020, 1 ,3),
                                    end_date=date(2020, 1, 3),
                                    user_agent="Name (email)")
            daily_filings.save("/path/to/directory")

            # limit which quarterly filings to use - saves only form 4 filings
            limit_to_form4 = lambda f: f.form_type.lower() == "4"
            daily_filings_limited = filings(start_date=date(2020, 1 ,3),
                                            end_date=date(2020, 1, 3),
                                            user_agent="Name (email)",
                                            entry_filter=limit_to_form4)
            daily_filings_limited.save("/path/to/other/directory")


        For getting filings from a specific quarter, the function call would look like this:


        .. code-block:: python

            from datetime import date
            from secedgar import filings

            # all quarterly filings
            quarterly_filings = filings(start_date=date(2020, 1 ,1),
                                        end_date=date(2020, 3, 31),
                                        user_agent="Name (email)")
            quarterly_filings.save("/path/to/directory")

            # limit which quarterly filings to use
            # saves only 10-K and 10-Q filings from quarter
            limit_to_10k_10q = lambda f: f.form_type.lower() in ("10-k", "10-q")
            quarterly_filings_limited = filings(start_date=date(2020, 1 ,1),
                                                end_date=date(2020, 3, 31),
                                                user_agent="Name (email)",
                                                entry_filter=limit_to_10k_10q)
            quarterly_filings_limited.save("/path/to/other/directory")

    """
    if filing_type is not None and not isinstance(filing_type, FilingType):
        raise FilingTypeError

    if cik_lookup:
        return CompanyFilings(
            cik_lookup,
            filing_type=filing_type,
            user_agent=user_agent,
            start_date=start_date,
            end_date=end_date,
            count=count,
            client=client,
            **kwargs
        )
    # Define entry filter as original
    _entry_filter = entry_filter

    if filing_type is not None:
        # If filing type also given, add filing types to existing entry filter
        def _entry_filter(x):  # noqa
            return x.form_type == filing_type and entry_filter(x)

    if count is not None:
        raise NotImplementedError(
            "Count has not yet been implemented for Daily, quarterly & Combo Filings."
        )

    if (end_date is None or end_date == start_date) and isinstance(
            start_date, date):
        return DailyFilings(date=start_date,
                            user_agent=user_agent,
                            client=client,
                            entry_filter=_entry_filter,
                            **kwargs)

    if isinstance(start_date, date) and isinstance(end_date, date):
        current_quarter = get_quarter(start_date)
        current_year = start_date.year
        start_quarter_date = date(current_year, get_month(current_quarter), 1)
        next_year, next_quarter = add_quarter(current_year, current_quarter)
        end_quarter_date = date(next_year, get_month(next_quarter),
                                1) - timedelta(days=1)
        if start_quarter_date == start_date and end_date == end_quarter_date:
            return QuarterlyFilings(year=current_year,
                                    quarter=current_quarter,
                                    client=client,
                                    user_agent=user_agent,
                                    entry_filter=_entry_filter,
                                    **kwargs)
        return ComboFilings(start_date=start_date,
                            end_date=end_date,
                            user_agent=user_agent,
                            client=client,
                            entry_filter=_entry_filter,
                            **kwargs)

    raise ValueError(
        """Invalid arguments. You must provide 'cik_lookup' OR 'start_date' \
OR ('start_date' and 'end_date').""")



# Tests if filings are correctly received from EDGAR
import datetime
import os

import pytest

from secedgar.cik_lookup import CIKLookup
from secedgar.client import NetworkClient
from secedgar.core import CompanyFilings, FilingType
from secedgar.exceptions import FilingTypeError, NoFilingsError
from secedgar.tests.utils import MockResponse


@pytest.fixture
def mock_cik_validator_get_single_cik(monkeypatch):
    """Mocks response for getting a single CIK."""
    monkeypatch.setattr(CIKLookup, "get_ciks",
                        lambda *args, **kwargs: {"aapl": "0000320193"})


@pytest.fixture(scope="module")
def mock_cik_validator_get_multiple_ciks(monkeymodule):
    """Mocks response for getting a single CIK."""
    monkeymodule.setattr(
        CIKLookup, "get_ciks", lambda *args: {
            "aapl": "0000320193",
            "msft": "1234",
            "amzn": "5678"
        })


@pytest.fixture(scope="module")
def mock_single_cik_not_found(monkeymodule):
    """NetworkClient get_response method will return html with CIK not found message."""
    monkeymodule.setattr(
        NetworkClient, "get_response",
        MockResponse(datapath_args=["CIK", "cik_not_found.html"]))


@pytest.fixture(scope="module")
def mock_single_cik_filing(monkeymodule):
    """Returns mock response of filinghrefs for getting filing URLs."""
    monkeymodule.setattr(
        NetworkClient, "get_response",
        MockResponse(datapath_args=["filings", "aapl_10q_filings.xml"]))


class MockSingleCIKFilingLimitedResponses:

    def __init__(self, num_responses):
        self._called_count = 0
        self._num_responses = num_responses

    def __call__(self, *args):
        if self._called_count * 10 < self._num_responses:
            self._called_count += 1
            return MockResponse(
                datapath_args=["filings", "aapl_10q_filings.xml"])
        else:
            return MockResponse(content=bytes("", "utf-8"))


# FIXME: This may not be working as expected. Need to look into this more.
@pytest.fixture
def mock_single_cik_filing_limited_responses(monkeypatch):
    """Mocks when only a limited number of filings are available.

    Should be reset with each function run, since calls left decreases
    after each call."""
    mock_limited_responses = MockSingleCIKFilingLimitedResponses(
        num_responses=10)
    monkeypatch.setattr(NetworkClient, "get_response", mock_limited_responses)


class TestCompanyFilings:
    valid_dates = [
        datetime.datetime(2020, 1, 1),
        datetime.datetime(2020, 2, 1),
        datetime.datetime(2020, 3, 1),
        datetime.datetime(2020, 4, 1),
        datetime.datetime(2020, 5, 1),
        "20200101",
        20200101,
        None
    ]
    bad_dates = [
        1,
        2020010101,
        "2020010101",
        "2020",
        "0102"
    ]

    class TestCompanyFilingsClient:

        def test_user_agent_client_none(self):
            with pytest.raises(TypeError):
                _ = CompanyFilings(cik_lookup="aapl",
                                   filing_type=FilingType.FILING_10Q,
                                   user_agent=None,
                                   client=None)

        def test_user_agent_set_to_client(self, mock_user_agent):
            aapl = CompanyFilings(cik_lookup="aapl",
                                  filing_type=FilingType.FILING_10Q,
                                  user_agent=mock_user_agent)
            assert aapl.client.user_agent == mock_user_agent

        def test_client_property_set(self, mock_user_agent):
            aapl = CompanyFilings(cik_lookup="aapl",
                                  filing_type=FilingType.FILING_10Q,
                                  client=NetworkClient(user_agent=mock_user_agent))
            assert aapl.client.user_agent == mock_user_agent

    def test_count_returns_exact(self, mock_user_agent, mock_cik_validator_get_single_cik,
                                 mock_single_cik_filing):
        count = 10
        aapl = CompanyFilings(user_agent=mock_user_agent,
                              cik_lookup="aapl",
                              filing_type=FilingType.FILING_10Q,
                              count=count)
        urls = aapl.get_urls()["aapl"]
        assert len(urls) == count, """Count should return exact number of filings.
                                 Got {0}, but expected {1} URLs.""".format(urls, count)

    @pytest.mark.parametrize("count", [None, 5, 10, 15, 27, 33])
    def test_count_setter_on_init(self, mock_user_agent, count):
        filing = CompanyFilings(user_agent=mock_user_agent,
                                cik_lookup="aapl",
                                filing_type=FilingType.FILING_10Q,
                                count=count)
        assert filing.count == count

    @pytest.mark.parametrize("start_date", valid_dates)
    def test_good_start_date_setter_on_init(self, start_date, mock_user_agent):
        filing = CompanyFilings(
            cik_lookup="aapl",
            filing_type=FilingType.FILING_10Q,
            start_date=start_date,
            user_agent=mock_user_agent)
        assert filing.start_date == start_date

    @pytest.mark.parametrize("bad_start_date", bad_dates)
    def test_bad_start_date_setter_on_init(self, mock_user_agent, bad_start_date):
        with pytest.raises(TypeError):
            CompanyFilings(user_agent=mock_user_agent,
                           cik_lookup="aapl",
                           filing_type=FilingType.FILING_10Q,
                           start_date=bad_start_date)

    @pytest.mark.parametrize("end_date", valid_dates)
    def test_good_end_date_setter_on_init(self, end_date, mock_user_agent):
        filing = CompanyFilings(
            cik_lookup="aapl",
            filing_type=FilingType.FILING_10Q,
            end_date=end_date,
            user_agent=mock_user_agent)
        assert filing.end_date == end_date

    @pytest.mark.parametrize("bad_end_date", bad_dates)
    def test_bad_end_date_setter_on_init(self, mock_user_agent, bad_end_date):
        with pytest.raises(TypeError):
            CompanyFilings(user_agent=mock_user_agent,
                           cik_lookup="aapl",
                           filing_type=FilingType.FILING_10Q,
                           end_date=bad_end_date)

    @pytest.mark.parametrize("count,expected_error", [(-1, ValueError),
                                                      (0, ValueError),
                                                      (0.0, TypeError),
                                                      (1.0, TypeError),
                                                      ("1", TypeError)])
    def test_count_setter_bad_values(self, mock_user_agent, count, expected_error):
        with pytest.raises(expected_error):
            CompanyFilings(user_agent=mock_user_agent,
                           cik_lookup="aapl",
                           filing_type=FilingType.FILING_10Q,
                           count=count)

    def test_date_is_sanitized(self, mock_user_agent):
        start_date = datetime.datetime(2012, 3, 1)
        end_date = datetime.datetime(2015, 1, 1)
        aapl = CompanyFilings(user_agent=mock_user_agent,
                              cik_lookup="aapl",
                              filing_type=FilingType.FILING_10Q,
                              count=10,
                              start_date=start_date,
                              end_date=end_date)
        assert aapl.params["dateb"] == "20150101"
        assert aapl.params["datea"] == "20120301"
        assert aapl.start_date == datetime.datetime(2012, 3, 1)
        assert aapl.end_date == datetime.datetime(2015, 1, 1)

    def test_date_is_sanitized_when_changed(self, mock_user_agent):
        aapl = CompanyFilings(user_agent=mock_user_agent,
                              cik_lookup="aapl",
                              filing_type=FilingType.FILING_10Q,
                              count=10,
                              start_date="20150101")
        assert aapl.start_date == "20150101"
        aapl.start_date = datetime.datetime(2010, 1, 1)
        assert aapl.start_date == datetime.datetime(2010, 1, 1)
        assert aapl.params["datea"] == "20100101"

    @pytest.mark.parametrize("date,expected",
                             [("20120101", "20120101"), (20120101, 20120101),
                              (datetime.datetime(2012, 1, 1), "20120101")])
    def test_end_date_setter(self, mock_user_agent, date, expected):
        f = CompanyFilings(cik_lookup="aapl",
                           user_agent=mock_user_agent,
                           filing_type=FilingType.FILING_10Q,
                           start_date=datetime.datetime(2010, 1, 1),
                           end_date=datetime.datetime(2015, 1, 1))
        f.end_date = date
        assert f.end_date == date and f.params.get("dateb") == expected

    def test_txt_urls(self, mock_user_agent,
                      mock_cik_validator_get_single_cik,
                      mock_single_cik_filing):
        aapl = CompanyFilings(user_agent=mock_user_agent,
                              cik_lookup="aapl",
                              filing_type=FilingType.FILING_10Q,
                              count=10)
        first_txt_url = aapl.get_urls()["aapl"][0]
        assert first_txt_url.split(".")[-1] == "txt"

    @pytest.mark.smoke
    def test_txt_urls_smoke(self, real_test_client):
        aapl = CompanyFilings(client=real_test_client,
                              cik_lookup="aapl",
                              filing_type=FilingType.FILING_10Q,
                              count=10)
        first_txt_url = aapl.get_urls()["aapl"][0]
        assert first_txt_url.split(".")[-1] == "txt"

    @pytest.mark.smoke
    def test_no_json_decode_error(self, mock_user_agent):
        from json import JSONDecodeError

        from secedgar import CompanyFilings, FilingType

        my_filings = CompanyFilings(cik_lookup='aapl',
                                    filing_type=FilingType.FILING_10Q,
                                    count=15,
                                    user_agent=mock_user_agent)
        try:
            my_filings.get_urls()
        except JSONDecodeError:
            pytest.fail("Received JSONDecodeError")

    @pytest.mark.parametrize("new_filing_type",
                             (FilingType.FILING_10K,
                              FilingType.FILING_8K,
                              FilingType.FILING_13FHR,
                              FilingType.FILING_SD,
                              None))
    def test_filing_type_setter(self, mock_user_agent, new_filing_type):
        f = CompanyFilings(user_agent=mock_user_agent,
                           cik_lookup="aapl",
                           filing_type=FilingType.FILING_10Q)
        f.filing_type = new_filing_type
        assert f.filing_type == new_filing_type

    @pytest.mark.parametrize("bad_filing_type",
                             ("10-k", "10k", "10-q", "10q", 123))
    def test_bad_filing_type_setter(self, mock_user_agent, bad_filing_type):
        f = CompanyFilings(user_agent=mock_user_agent,
                           cik_lookup="aapl",
                           filing_type=FilingType.FILING_10Q)
        with pytest.raises(FilingTypeError):
            f.filing_type = bad_filing_type

    @pytest.mark.parametrize("bad_filing_type",
                             ("10-j", "10-k", "ssd", "invalid", 1))
    def test_invalid_filing_type_types(self, mock_user_agent, bad_filing_type):
        with pytest.raises(FilingTypeError):
            CompanyFilings(user_agent=mock_user_agent,
                           cik_lookup="0000320193",
                           filing_type=bad_filing_type)

    @pytest.mark.parametrize("bad_cik_lookup", (1234567891011, 12345, 123.0))
    def test_validate_cik_type_inside_filing(self, mock_user_agent, bad_cik_lookup):
        with pytest.raises(TypeError):
            CompanyFilings(user_agent=mock_user_agent,
                           cik_lookup=bad_cik_lookup,
                           filing_type=FilingType.FILING_10K)

    @pytest.mark.parametrize("no_urls", ({}, {
        "aapl": [],
        "fb": [],
        "msft": []
    }))
    def test_save_no_filings_raises_error(self, tmp_data_directory, monkeypatch,
                                          mock_user_agent, no_urls):
        monkeypatch.setattr(CompanyFilings, "get_urls", lambda x: no_urls)
        f = CompanyFilings(user_agent=mock_user_agent,
                           cik_lookup="aapl", filing_type=FilingType.FILING_10K)
        with pytest.raises(NoFilingsError):
            f.save(tmp_data_directory)

    def test_filing_save_multiple_ciks(self, tmp_data_directory,
                                       mock_user_agent,
                                       mock_cik_validator_get_multiple_ciks,
                                       mock_single_cik_filing,
                                       mock_filing_response):
        f = CompanyFilings(["aapl", "amzn", "msft"],
                           FilingType.FILING_10Q,
                           user_agent=mock_user_agent,
                           count=3)
        f.save(tmp_data_directory)
        assert len(os.listdir(tmp_data_directory)) > 0

    @pytest.mark.parametrize(
        "match_format",
        [
            "EXACT",
            "AMEND",
            "ALL"
        ]
    )
    def test_match_format_good(self, mock_user_agent, match_format):
        f = CompanyFilings(["aapl", "amzn", "msft"],
                           FilingType.FILING_10Q,
                           user_agent=mock_user_agent,
                           count=3,
                           match_format=match_format)

        assert f.match_format == match_format

    @pytest.mark.parametrize(
        "match_format",
        [
            "exact",
            "amend",
            "all",
            "none",
            None,
            True,
            False,
            1,
            0
        ]
    )
    def test_match_format_bad(self, mock_user_agent, match_format):
        with pytest.raises(ValueError):
            CompanyFilings(["aapl", "amzn", "msft"],
                           FilingType.FILING_10Q,
                           user_agent=mock_user_agent,
                           count=3,
                           match_format=match_format)

    @pytest.mark.smoke
    def test_filing_save_multiple_ciks_smoke(self, tmp_data_directory,
                                             real_test_client):
        f = CompanyFilings(["aapl", "amzn", "msft"],
                           filing_type=FilingType.FILING_10Q,
                           client=real_test_client,
                           count=3)
        f.save(tmp_data_directory)
        assert len(os.listdir(tmp_data_directory)) > 0

    @pytest.mark.asyncio
    @pytest.mark.smoke
    def test_filing_company_none_filing_type(self, tmp_data_directory, real_test_client):
        f = CompanyFilings(["aapl", "msft", "amzn"],
                           filing_type=None,
                           client=real_test_client,
                           count=3)
        f.save(tmp_data_directory)
        assert len(os.listdir(tmp_data_directory)) > 0

    def test_filing_save_single_cik(self, tmp_data_directory,
                                    mock_user_agent,
                                    mock_cik_validator_get_single_cik,
                                    mock_single_cik_filing,
                                    mock_filing_response):
        f = CompanyFilings("aapl", FilingType.FILING_10Q, user_agent=mock_user_agent, count=3)
        f.save(tmp_data_directory)
        assert len(os.listdir(tmp_data_directory)) > 0

    @pytest.mark.smoke
    def test_filing_save_single_cik_smoke(self, tmp_data_directory,
                                          real_test_client):
        f = CompanyFilings("aapl", FilingType.FILING_10Q, client=real_test_client, count=3)
        f.save(tmp_data_directory)
        assert len(os.listdir(tmp_data_directory)) > 0

    def test_filing_get_urls_returns_single_list_of_urls(
            self, mock_user_agent,
            mock_cik_validator_get_multiple_ciks, mock_single_cik_filing):
        # Uses same response for filing links (will all be filings for aapl)
        f = CompanyFilings(cik_lookup=["aapl", "msft", "amzn"],
                           user_agent=mock_user_agent,
                           filing_type=FilingType.FILING_10Q,
                           count=5)
        assert all(
            len(f.get_urls().get(key)) == 5 for key in f.get_urls().keys())

    @pytest.mark.parametrize("count", [10, 25, 30])
    def test_filing_returns_correct_number_of_urls(
            self, count, mock_user_agent, mock_cik_validator_get_multiple_ciks,
            mock_single_cik_filing):
        # Uses same response for filing links (will all be filings for aapl)
        f = CompanyFilings(cik_lookup=["aapl", "msft", "amzn"],
                           filing_type=FilingType.FILING_10Q,
                           user_agent=mock_user_agent,
                           count=count,
                           client=NetworkClient(user_agent=mock_user_agent, batch_size=10))
        assert all(
            len(f.get_urls().get(key)) == count for key in f.get_urls().keys())

    @pytest.mark.parametrize("count,raises_error", [(5, False),
                                                    (10, False),
                                                    (20, True),
                                                    (30, True),
                                                    (40, True)])
    @pytest.mark.filterwarnings("ignore::DeprecationWarning")
    # For collections.abc warning 3.8+
    def test_filings_warning_lt_count(self, recwarn, count, raises_error,
                                      tmp_data_directory,
                                      mock_user_agent,
                                      mock_cik_validator_get_single_cik,
                                      mock_single_cik_filing_limited_responses,
                                      mock_filing_response):  # noqa:E501
        f = CompanyFilings(cik_lookup="aapl",
                           filing_type=FilingType.FILING_10Q,
                           count=count,
                           client=NetworkClient(batch_size=10, user_agent=mock_user_agent))
        f.save(tmp_data_directory)
        if raises_error:
            w = recwarn.pop(UserWarning)
            assert issubclass(w.category, UserWarning)
        else:
            try:
                w = recwarn.pop(UserWarning)
                # Allow XMLParsedAsHTMLWarning, but don't allow others
                if w and w._category_name == "XMLParsedAsHTMLWarning":
                    pass
                else:
                    pytest.fail("Expected no UserWarning, but received one.")
            # Should raise assertion error since no UserWarning should be found
            except AssertionError:
                pass

    @pytest.mark.smoke
    @pytest.mark.slow
    def test_filing_simple_example_smoke(self, tmp_data_directory,
                                         mock_user_agent):
        my_filings = CompanyFilings(cik_lookup="IBM",
                                    filing_type=FilingType.FILING_10Q,
                                    user_agent=mock_user_agent,
                                    count=3)
        my_filings.save(tmp_data_directory)
        assert len(os.listdir(tmp_data_directory)) > 0, "No file or directory created after save."

    def test__filter_filing_links(self, mock_user_agent, mock_single_cik_filing):
        f = CompanyFilings(cik_lookup="aapl",
                           filing_type=FilingType.FILING_10Q,
                           user_agent=mock_user_agent)
        data = f.client.get_soup("", {})  # Get mock data from mock_single_cik_filing
        links = f._filter_filing_links(data)
        assert len(links) == 10
        assert all(["BAD_LINK" not in link for link in links])

    def test_same_urls_fetched(self, mock_user_agent, mock_single_cik_filing):
        # mock_single_filing_cik has more than 10 URLs
        # using count = 5 should help test whether the same URLs
        # are fetched each time
        f = CompanyFilings(cik_lookup="aapl",
                           filing_type=FilingType.FILING_10Q,
                           user_agent=mock_user_agent,
                           count=5)
        first_urls = f.get_urls()
        second_urls = f.get_urls()
        assert all(f == s for f, s in zip(first_urls, second_urls))

    @pytest.mark.parametrize(
        "bad_ownership",
        [
            "notright",
            "_exclude",
            "_include",
            "notvalid",
            1,
            True,
            False
        ]
    )
    def test_ownership(self, bad_ownership, mock_user_agent):
        with pytest.raises(ValueError):
            CompanyFilings(
                cik_lookup="aapl",
                filing_type=FilingType.FILING_10Q,
                user_agent=mock_user_agent,
                ownership=bad_ownership
            )

    def test_good_ownership(self, mock_user_agent):
        shared_params = dict(
            cik_lookup="aapl",
            filing_type=FilingType.FILING_10Q,
            user_agent=mock_user_agent,
        )
        f_include = CompanyFilings(
            **shared_params,
            ownership="include"
        )
        f_exclude = CompanyFilings(
            **shared_params,
            ownership="exclude"
        )
        assert f_include.ownership == "include"
        assert f_exclude.ownership == "exclude"

        # Change ownership type
        f_include.ownership = "exclude"
        assert f_include.ownership == "exclude"

    def test_start_date_change_to_none(self, mock_user_agent):
        start_date = datetime.date(2020, 1, 1)
        f = CompanyFilings(
            cik_lookup="aapl",
            filing_type=FilingType.FILING_10Q,
            user_agent=mock_user_agent,
            start_date=start_date
        )
        assert f.start_date == start_date
        assert f.params["datea"] == "20200101"
        f.start_date = None
        assert f.start_date is None
        assert "datea" not in f.params

    def test_end_date_change_to_none(self, mock_user_agent):
        end_date = datetime.date(2020, 1, 1)
        f = CompanyFilings(
            cik_lookup="aapl",
            filing_type=FilingType.FILING_10Q,
            user_agent=mock_user_agent,
            end_date=end_date
        )
        assert f.end_date == end_date
        assert f.params["dateb"] == "20200101"
        f.end_date = None
        assert f.end_date is None
        assert "dateb" not in f.params



import os
from datetime import date

import pytest
from secedgar.core.quarterly import QuarterlyFilings
from secedgar.tests.utils import MockResponse


@pytest.fixture(scope="module")
def mock_quarterly_quarter_directory(monkeymodule):
    """Mock directory of all filings for quarter.

    Use for QuarterlyFilings object.
    """
    monkeymodule.setattr(
        QuarterlyFilings, "_get_listings_directory",
        MockResponse(
            datapath_args=["filings", "master", "master_index_1993_QTR4.html"]))


@pytest.fixture
def mock_master_idx_file(monkeypatch):
    monkeypatch.setattr(
        QuarterlyFilings, "_get_master_idx_file", lambda *args: MockResponse(
            datapath_args=["filings", "master", "master.idx"]).text)


class TestQuarterly:

    @pytest.mark.parametrize("bad_year,expected_error", [
        (-1, ValueError),
        (0.0, TypeError),
        (1990, ValueError),
        (1991, ValueError),
        (1992, ValueError),
        ("1993", TypeError),
        ("1993.0", TypeError),
    ])
    def test_bad_year(self, mock_user_agent, bad_year, expected_error):
        with pytest.raises(expected_error):
            _ = QuarterlyFilings(year=bad_year, quarter=1, user_agent=mock_user_agent)

    def test_good_year(self, mock_user_agent):
        for year in range(1993, date.today().year + 1):
            mf = QuarterlyFilings(year=year, quarter=1, user_agent=mock_user_agent)
            assert mf.year == year

    @pytest.mark.parametrize("bad_quarter,expected_error", [(0.0, TypeError),
                                                            (1.0, TypeError),
                                                            ("1", TypeError),
                                                            ("1.0", TypeError),
                                                            (0, ValueError),
                                                            (5, ValueError),
                                                            (6, ValueError),
                                                            (2020, ValueError)])
    def test_bad_quarter(self, mock_user_agent, bad_quarter, expected_error):
        with pytest.raises(expected_error):
            _ = QuarterlyFilings(year=2020, quarter=bad_quarter, user_agent=mock_user_agent)

    def test_good_quarters(self, mock_user_agent):
        for quarter in range(1, 5):
            mf = QuarterlyFilings(year=2019, quarter=quarter, user_agent=mock_user_agent)
            assert mf.quarter == quarter

    @pytest.mark.parametrize("year,quarter", [(2018, 1), (2019, 2), (2020, 3)])
    def test_idx_filename_is_always_the_same(self, mock_user_agent, year, quarter):
        mf = QuarterlyFilings(year=year, quarter=quarter, user_agent=mock_user_agent)
        assert mf.idx_filename == "master.idx"

    def test_always_false_entry_filter(self, mock_user_agent, mock_master_idx_file):
        quarterly_filing = QuarterlyFilings(year=1993,
                                            quarter=4,
                                            entry_filter=lambda _: False,
                                            user_agent=mock_user_agent)
        urls = quarterly_filing.get_urls()
        assert len(urls) == 0

    @pytest.mark.parametrize("subdir,file", [
        ("1095785", "9999999997-02-056978.txt"),
        ("11860", "0000011860-94-000005.txt"),
        ("17206", "0000017206-94-000007.txt"),
        ("205239", "0000205239-94-000003.txt"),
        ("20762", "0000950131-94-000025.txt"),
    ])
    def test_save(self, tmp_data_directory, mock_user_agent, mock_filing_data,
                  mock_quarterly_quarter_directory, mock_master_idx_file,
                  mock_filing_response, subdir, file):
        quarterly_filing = QuarterlyFilings(year=1993, quarter=4, user_agent=mock_user_agent)
        quarterly_filing.save(tmp_data_directory)
        subdir = os.path.join("1993", "QTR4", subdir)
        path_to_check = os.path.join(tmp_data_directory, subdir, file)
        assert os.path.exists(path_to_check)

    @pytest.mark.parametrize("original_path,clean_path", [
        ("Apple Inc.", "Apple_Inc"),
        ("Microsoft Corporation", "Microsoft_Corporation"),
        ("Bed, Bath, & Beyond", "Bed_Bath__Beyond"),
        ("Company with \\lots\\ of /slashes/", "Company_with_lots_of_slashes")
    ])
    def test_clean_path(self, mock_user_agent, original_path, clean_path):
        quarterly_filing = QuarterlyFilings(year=2000, quarter=1, user_agent=mock_user_agent)
        assert quarterly_filing.clean_directory_path(original_path) == clean_path

    def test_user_agent_passed_to_client(self, mock_user_agent):
        quarterly = QuarterlyFilings(year=2020, quarter=1, user_agent=mock_user_agent)
        assert quarterly.client.user_agent == mock_user_agent

    def test_error_raised_no_user_agent(self):
        with pytest.raises(TypeError):
            _ = QuarterlyFilings(year=2020, quarter=1)



from enum import Enum


class FilingType(Enum):
    """Available filing types to be used when creating Filing object.

    .. versionadded:: 0.1.5

    """
    FILING_1 = '1'
    FILING_10 = '10'
    FILING_1012B = '10-12B'
    FILING_1012B_AMEND = '10-12B/A'
    FILING_1012G = '10-12G'
    FILING_1012G_AMEND = '10-12G/A'
    FILING_10C = '10-c'
    FILING_10D = '10-D'
    FILING_10D_AMEND = '10-D/A'
    FILING_10K = '10-K'
    FILING_10K405 = '10-k405'
    FILING_10KSB = '10ksb'
    FILING_10KSB40 = '10ksb40'
    FILING_10KSB405 = '10ksb405'
    FILING_10KT = '10-KT'
    FILING_10KT405 = '10kt405'
    FILING_10KT_AMEND = '10-KT/A'
    FILING_10K_AMEND = '10-K/A'
    FILING_10M = '10-m'
    FILING_10Q = '10-Q'
    FILING_10QSB = '10qsb'
    FILING_10QT = '10-QT'
    FILING_10QT_AMEND = '10-QT/A'
    FILING_10Q_AMEND = '10-Q/A'
    FILING_10SB12B = '10sb12b'
    FILING_10SB12G = '10sb12g'
    FILING_11K = '11-K'
    FILING_11KT = '11-KT'
    FILING_11KT_AMEND = '11-KT/A'
    FILING_11K_AMEND = '11-K/A'
    FILING_12B25 = '12b-25'
    FILING_12G2 = '12g-2'
    FILING_12G32A = '12g3-2a'
    FILING_12G32B = '12g3-2b'
    FILING_12G32BR = '12g32br'
    FILING_13F = '13f'
    FILING_13FCONP = '13fconp'
    FILING_13FE = '13f-e'
    FILING_13FHR = '13F-HR'
    FILING_13FHR_AMEND = '13F-HR/A'
    FILING_13FNT = '13F-NT'
    FILING_13FNT_AMEND = '13F-NT/A'
    FILING_13H = '13H'
    FILING_13HA = '13H-A'
    FILING_13HI = '13H-I'
    FILING_13HQ = '13H-Q'
    FILING_13HR = '13H-R'
    FILING_13HT = '13H-T'
    FILING_144 = '144'
    FILING_144_AMEND = '144/A'
    FILING_15 = '15'
    FILING_1512B = '15-12B'
    FILING_1512B_AMEND = '15-12B/A'
    FILING_1512G = '15-12G'
    FILING_1512G_AMEND = '15-12G/A'
    FILING_1515D = '15-15D'
    FILING_1515D_AMEND = '15-15D/A'
    FILING_15F = '15f'
    FILING_15F12B = '15F-12B'
    FILING_15F12B_AMEND = '15F-12B/A'
    FILING_15F12G = '15F-12G'
    FILING_15F12G_AMEND = '15F-12G/A'
    FILING_15F15D = '15F-15D'
    FILING_15F15D_AMEND = '15F-15D/A'
    FILING_17H = '17-h'
    FILING_17HACON = '17HACON'
    FILING_17HACON_AMEND = '17HACON/A'
    FILING_17HQCON = '17HQCON'
    FILING_17HQCON_AMEND = '17HQCON/A'
    FILING_18 = '18'
    FILING_1812B = '18-12B'
    FILING_1812B_AMEND = '18-12B/A'
    FILING_1812G = '18-12G'
    FILING_1812G_AMEND = '18-12G/A'
    FILING_18K = '18-K'
    FILING_18K_AMEND = '18-K/A'
    FILING_19B = '19-b'
    FILING_19B4 = '19b-4'
    FILING_19B4E = '19b-4(e)'
    FILING_19B7 = '19b-7'
    FILING_1A = '1-A'
    FILING_1APOS = '1-A POS'
    FILING_1AW = '1-A-W'
    FILING_1AW_AMEND = '1-A-W/A'
    FILING_1A_AMEND = '1-A/A'
    FILING_1E = '1-E'
    FILING_1EAD = '1-E AD'
    FILING_1EAD_AMEND = '1-E AD/A'
    FILING_1E_AMEND = '1-E/A'
    FILING_1K = '1-K'
    FILING_1K_AMEND = '1-K/A'
    FILING_1N = '1-n'
    FILING_1SA = '1-SA'
    FILING_1SA_AMEND = '1-SA/A'
    FILING_1U = '1-U'
    FILING_1U_AMEND = '1-U/A'
    FILING_1Z = '1-Z'
    FILING_1ZW = '1-Z-W'
    FILING_1ZW_AMEND = '1-Z-W/A'
    FILING_1Z_AMEND = '1-Z/A'
    FILING_20F = '20-F'
    FILING_20FR = '20-fr'
    FILING_20FR12B = '20FR12B'
    FILING_20FR12B_AMEND = '20FR12B/A'
    FILING_20FR12G = '20FR12G'
    FILING_20FR12G_AMEND = '20FR12G/A'
    FILING_20F_AMEND = '20-F/A'
    FILING_24F1 = '24f-1'
    FILING_24F2 = '24f-2'
    FILING_24F2EL = '24f-2el'
    FILING_24F2NT = '24F-2NT'
    FILING_24F2NT_AMEND = '24F-2NT/A'
    FILING_24F2TM = '24f-2tm'
    FILING_25 = '25'
    FILING_253G1 = '253G1'
    FILING_253G2 = '253G2'
    FILING_253G3 = '253G3'
    FILING_253G4 = '253G4'
    FILING_25NSE = '25-NSE'
    FILING_25NSE_AMEND = '25-NSE/A'
    FILING_25_AMEND = '25/A'
    FILING_26 = '26'
    FILING_27 = '27'
    FILING_28 = '28'
    FILING_2A = '2-a'
    FILING_2AF = '2-af'
    FILING_2E = '2-E'
    FILING_2E_AMEND = '2-E/A'
    FILING_3 = '3'
    FILING_305B2 = '305B2'
    FILING_305B2_AMEND = '305B2/A'
    FILING_3412H = '34-12h'
    FILING_3436CF = '34-36cf'
    FILING_3436MR = '34-36mr'
    FILING_3511 = '35-11'
    FILING_352 = '35-2'
    FILING_353 = '35-3'
    FILING_357B = '35-7b'
    FILING_35APP = '35-app'
    FILING_35CERT = '35-cert'
    FILING_3910B2 = '39-10b2'
    FILING_39304C = '39-304c'
    FILING_39304D = '39-304d'
    FILING_39310B = '39-310b'
    FILING_3_AMEND = '3/A'
    FILING_4 = '4'
    FILING_4017F1 = '40-17F1'
    FILING_4017F1_AMEND = '40-17F1/A'
    FILING_4017F2 = '40-17F2'
    FILING_4017F2_AMEND = '40-17F2/A'
    FILING_4017G = '40-17G'
    FILING_4017GCS = '40-17GCS'
    FILING_4017GCS_AMEND = '4017GCS/A'
    FILING_4017G_AMEND = '40-17G/A'
    FILING_40202A = '40-202a'
    FILING_40203A = '40-203a'
    FILING_40205A = '40-205a'
    FILING_40205E = '40-205e'
    FILING_40206A = '40-206a'
    FILING_4024B2 = '40-24B2'
    FILING_4024B2_AMEND = '40-24B2/A'
    FILING_4033 = '40-33'
    FILING_4033_AMEND = '40-33/A'
    FILING_406B = '40-6B'
    FILING_406B_AMEND = '40-6B/A'
    FILING_406C = '40-6c'
    FILING_408B25 = '40-8B25'
    FILING_408F2 = '40-8F-2'
    FILING_408F2_AMEND = '40-8F-2/A'
    FILING_408FA = '40-8f-a'
    FILING_408FB = '40-8f-b'
    FILING_408FC = '40-8fc'
    FILING_408FL = '40-8f-l'
    FILING_408FM = '40-8f-m'
    FILING_40APP = '40-APP'
    FILING_40APP_AMEND = '40-APP/A'
    FILING_40F = '40-F'
    FILING_40FR12B = '40FR12B'
    FILING_40FR12B_AMEND = '40FR12B/A'
    FILING_40FR12G = '40FR12G'
    FILING_40FR12G_AMEND = '40FR12G/A'
    FILING_40F_AMEND = '40-F/A'
    FILING_40OIP = '40-OIP'
    FILING_40OIP_AMEND = '40-OIP/A'
    FILING_40RPT = '40-rpt'
    FILING_424A = '424A'
    FILING_424B1 = '424B1'
    FILING_424B2 = '424B2'
    FILING_424B3 = '424B3'
    FILING_424B4 = '424B4'
    FILING_424B5 = '424B5'
    FILING_424B7 = '424B7'
    FILING_424B8 = '424B8'
    FILING_424H = '424H'
    FILING_424H_AMEND = '424H/A'
    FILING_425 = '425'
    FILING_45B3 = '45b-3'
    FILING_485A24E = '485a24e'
    FILING_485A24F = '485a24f'
    FILING_485APOS = '485APOS'
    FILING_485B24E = '485b24e'
    FILING_485B24F = '485b24f'
    FILING_485BPOS = '485BPOS'
    FILING_485BXT = '485BXT'
    FILING_485BXTF = '485bxtf'
    FILING_486A24E = '486a24e'
    FILING_486APOS = '486APOS'
    FILING_486B24E = '486b24e'
    FILING_486BPOS = '486BPOS'
    FILING_486BXT = '486BXT'
    FILING_487 = '487'
    FILING_497 = '497'
    FILING_497AD = '497AD'
    FILING_497H2 = '497H2'
    FILING_497J = '497J'
    FILING_497K = '497K'
    FILING_497K1 = '497k1'
    FILING_497K2 = '497k2'
    FILING_497K3A = '497k3a'
    FILING_497K3B = '497k3b'
    FILING_497VPI = '497VPI'
    FILING_497VPU = '497VPU'
    FILING_4_AMEND = '4/A'
    FILING_5 = '5'
    FILING_5_AMEND = '5/A'
    FILING_6B_NTC = '6b ntc'
    FILING_6B_ORDR = '6b ordr'
    FILING_6K = '6-K'
    FILING_6K_AMEND = '6-K/A'
    FILING_7A = '7-a'
    FILING_8A12B = '8-A12B'
    FILING_8A12BEF = '8a12bef'
    FILING_8A12BT = '8a12bt'
    FILING_8A12B_AMEND = '8-A12B/A'
    FILING_8A12G = '8-A12G'
    FILING_8A12G_AMEND = '8-A12G/A'
    FILING_8B12B = '8-b12b'
    FILING_8B12G = '8-b12g'
    FILING_8F2_NTC = '8f-2 ntc'
    FILING_8F2_ORDR = '8f-2 ordr'
    FILING_8K = '8-K'
    FILING_8K12B = '8-K12B'
    FILING_8K12B_AMEND = '8-K12B/A'
    FILING_8K12G3 = '8-K12G3'
    FILING_8K12G3_AMEND = '8-K12G3/A'
    FILING_8K15D5 = '8-K15D5'
    FILING_8K15D5_AMEND = '8-K15D5/A'
    FILING_8K_AMEND = '8-K/A'
    FILING_8M = '8-m'
    FILING_9M = '9-m'
    FILING_ABS15G = 'ABS-15G'
    FILING_ABS15G_AMEND = 'ABS-15G/A'
    FILING_ABSEE = 'ABS-EE'
    FILING_ABSEE_AMEND = 'ABS-EE/A'
    FILING_ADB = 'adb'
    FILING_ADNMTL = 'adn-mtl'
    FILING_ADV = 'adv'
    FILING_ADVCO = 'advco'
    FILING_ADVE = 'adv-e'
    FILING_ADVH = 'adv-h'
    FILING_ADVHC = 'adv-h-c'
    FILING_ADVHT = 'adv-h-t'
    FILING_ADVNR = 'adv-nr'
    FILING_ADVW = 'adv-w'
    FILING_AFDB = 'afdb'
    FILING_ANNLRPT = 'ANNLRPT'
    FILING_ANNLRPT_AMEND = 'ANNLRPT/A'
    FILING_APPWD = 'APP WD'
    FILING_APPWDAPPWD_AMEND = 'APP WD APP WD/A'
    FILING_APPWD_AMEND = 'APP WD/A'
    FILING_APP_NTC = 'app ntc'
    FILING_APP_ORDR = 'app ordr'
    FILING_APP_WD = 'app wd'
    FILING_ARS = 'ARS'
    FILING_ARS_AMEND = 'ARS/A'
    FILING_ATS = 'ats'
    FILING_ATSN = 'ATS-N'
    FILING_ATSNC = 'ATS-N-C'
    FILING_ATSNCA = 'ATS-N/CA'
    FILING_ATSNMA = 'ATS-N/MA'
    FILING_ATSNOFA = 'ATS-N/OFA'
    FILING_ATSNUA = 'ATS-N/UA'
    FILING_ATSNW = 'ATS-N-W'
    FILING_ATSN_A_ORDR_INEFF = 'ats-n/a ordr ineff'
    FILING_ATSN_CA = 'ats-n/ca'
    FILING_ATSN_MA = 'ats-n/ma'
    FILING_ATSN_MA_CP = 'ats-n/ma cp'
    FILING_ATSN_OFA = 'ats-n/ofa'
    FILING_ATSN_ORDR_INEFF = 'ats-n ordr ineff'
    FILING_ATSN_ORDR_LTD_PN = 'ats-n ordr ltd opn'
    FILING_ATSN_ORDR_REVK = 'ats-n ordr revk'
    FILING_ATSN_ORDR_SUSP = 'ats-n ordr susp'
    FILING_ATSN_UA = 'ats-n/ua'
    FILING_ATSR = 'ats-r'
    FILING_AW = 'AW'
    FILING_AWWD = 'AW WD'
    FILING_AW_WD = 'aw wd'
    FILING_BD = 'bd'
    FILING_BDCO = 'bdco'
    FILING_BDN = 'bd-n'
    FILING_BDW = 'bdw'
    FILING_BULK = 'BULK'
    FILING_BW2 = 'bw-2'
    FILING_BW3 = 'bw-3'
    FILING_C = 'C'
    FILING_CA1 = 'ca-1'
    FILING_CAR = 'C-AR'
    FILING_CARAW = 'C-AR/A-W'
    FILING_CARW = 'C-AR-W'
    FILING_CAR_AMEND = 'C-AR/A'
    FILING_CAW = 'C/A-W'
    FILING_CB = 'CB'
    FILING_CB_AMEND = 'CB/A'
    FILING_CERT = 'CERT'
    FILING_CERTAMX = 'certamx'
    FILING_CERTARCA = 'certarca'
    FILING_CERTBSE = 'certbse'
    FILING_CERTCBO = 'certcbo'
    FILING_CERTCIN = 'certcin'
    FILING_CERTCSE = 'certcse'
    FILING_CERTISE = 'certise'
    FILING_CERTNAS = 'certnas'
    FILING_CERTNYS = 'certnys'
    FILING_CERTPAC = 'certpac'
    FILING_CERTPBS = 'certpbs'
    FILING_CFPORTAL = 'CFPORTAL'
    FILING_CFPORTALW = 'CFPORTAL-W'
    FILING_CFPORTAL_AMEND = 'CFPORTAL/A'
    FILING_CORRESP = 'CORRESP'
    FILING_CTR = 'C-TR'
    FILING_CTRW = 'C-TR-W'
    FILING_CT_ORDER = 'ct order'
    FILING_CU = 'C-U'
    FILING_CUSTODY = 'custody'
    FILING_CUW = 'C-U-W'
    FILING_CW = 'C-W'
    FILING_C_AMEND = 'C/A'
    FILING_C_ARW = 'c-ar-w'
    FILING_C_AR_AW = 'c-ar/a-w'
    FILING_C_AW = 'c/a-w'
    FILING_C_TRW = 'c-tr-w'
    FILING_C_UW = 'c-u-w'
    FILING_D = 'D'
    FILING_DEF13E3 = 'def13e3'
    FILING_DEF14A = 'DEF 14A'
    FILING_DEF14C = 'DEF 14C'
    FILING_DEFA14A = 'DEFA14A'
    FILING_DEFA14C = 'DEFA14C'
    FILING_DEFC14A = 'DEFC14A'
    FILING_DEFC14C = 'DEFC14C'
    FILING_DEFM14A = 'DEFM14A'
    FILING_DEFM14C = 'DEFM14C'
    FILING_DEFN14A = 'DEFN14A'
    FILING_DEFOC = 'def-oc'
    FILING_DEFR14A = 'DEFR14A'
    FILING_DEFR14C = 'DEFR14C'
    FILING_DEFS14A = 'defs14a'
    FILING_DEFS14C = 'defs14c'
    FILING_DEF_14A = 'def 14a'
    FILING_DEF_14C = 'def 14c'
    FILING_DELAM = 'DEL AM'
    FILING_DEL_AM = 'del am'
    FILING_DFAN14A = 'DFAN14A'
    FILING_DFRN14A = 'DFRN14A'
    FILING_DOS = 'DOS'
    FILING_DOSLTR = 'DOSLTR'
    FILING_DOS_AMEND = 'DOS/A'
    FILING_DRS = 'DRS'
    FILING_DRSLTR = 'DRSLTR'
    FILING_DRS_AMEND = 'DRS/A'
    FILING_DSTRBRPT = 'DSTRBRPT'
    FILING_DSTRBRPT_AMEND = 'DSTRBRPT/A'
    FILING_D_AMEND = 'D/A'
    FILING_EBRD = 'ebrd'
    FILING_EFFECT = 'effect'
    FILING_F1 = 'F-1'
    FILING_F10 = 'F-10'
    FILING_F10EF = 'F-10EF'
    FILING_F10MEF = 'f-10mef'
    FILING_F10POS = 'F-10POS'
    FILING_F10_AMEND = 'F-10/A'
    FILING_F1MEF = 'F-1MEF'
    FILING_F1_AMEND = 'F-1/A'
    FILING_F2 = 'f-2'
    FILING_F2D = 'f-2d'
    FILING_F2DPOS = 'f-2dpos'
    FILING_F2MEF = 'f-2mef'
    FILING_F3 = 'F-3'
    FILING_F3ASR = 'F-3ASR'
    FILING_F3D = 'F-3D'
    FILING_F3DPOS = 'F-3DPOS'
    FILING_F3MEF = 'F-3MEF'
    FILING_F3_AMEND = 'F-3/A'
    FILING_F4 = 'F-4'
    FILING_F4EF = 'F-4EF'
    FILING_F4MEF = 'F-4MEF'
    FILING_F4POS = 'F-4 POS'
    FILING_F4_AMEND = 'F-4/A'
    FILING_F4_POS = 'f-4 pos'
    FILING_F6 = 'F-6'
    FILING_F6EF = 'F-6EF'
    FILING_F6POS = 'F-6 POS'
    FILING_F6_AMEND = 'F-6/A'
    FILING_F6_POS = 'f-6 pos'
    FILING_F7 = 'F-7'
    FILING_F7POS = 'F-7 POS'
    FILING_F7_AMEND = 'F-7/A'
    FILING_F7_POS = 'f-7 pos'
    FILING_F8 = 'F-8'
    FILING_F80 = 'F-80'
    FILING_F80POS = 'F-80POS'
    FILING_F80_AMEND = 'F-80/A'
    FILING_F8POS = 'F-8 POS'
    FILING_F8_AMEND = 'F-8/A'
    FILING_F8_POS = 'f-8 pos'
    FILING_F9 = 'f-9'
    FILING_F9EF = 'f-9ef'
    FILING_F9MEF = 'f-9mef'
    FILING_F9_POS = 'f-9 pos'
    FILING_FN = 'F-N'
    FILING_FN_AMEND = 'F-N/A'
    FILING_FOCUSN = 'focusn'
    FILING_FWP = 'FWP'
    FILING_FX = 'F-X'
    FILING_FX_AMEND = 'F-X/A'
    FILING_G405 = 'g-405'
    FILING_GFIN = 'g-fin'
    FILING_GFINW = 'g-finw'
    FILING_HISTORY = 'history'
    FILING_IADB = 'iadb'
    FILING_IBRD = 'ibrd'
    FILING_ID = 'id'
    FILING_IDNEWCIK = 'id-newcik'
    FILING_IFC = 'ifc'
    FILING_IRANNOTICE = 'IRANNOTICE'
    FILING_MA = 'MA'
    FILING_MAA = 'MA-A'
    FILING_MAI = 'MA-I'
    FILING_MAI_AMEND = 'MA-I/A'
    FILING_MANR = 'ma-nr'
    FILING_MAW = 'MA-W'
    FILING_MA_AMEND = 'MA/A'
    FILING_MODULE = 'MODULE'
    FILING_MSD = 'msd'
    FILING_MSDCO = 'msdco'
    FILING_MSDW = 'msdw'
    FILING_N1 = 'n-1'
    FILING_N14 = 'N-14'
    FILING_N148C = 'N-14 8C'
    FILING_N148C_AMEND = 'N-14 8C/A'
    FILING_N14AE = 'n-14ae'
    FILING_N14AE24 = 'n14ae24'
    FILING_N14EL24 = 'n14el24'
    FILING_N14MEF = 'N-14MEF'
    FILING_N14_8C = 'n-14 8c'
    FILING_N14_AMEND = 'N-14/A'
    FILING_N17D1 = 'n-17d-1'
    FILING_N17F1 = 'n-17f-1'
    FILING_N17F2 = 'n-17f-2'
    FILING_N18F1 = 'N-18F1'
    FILING_N18F1_AMEND = 'N-18F1/A'
    FILING_N1A = 'N-1A'
    FILING_N1A_AMEND = 'N-1A/A'
    FILING_N1A_EL = 'n-1a el'
    FILING_N2 = 'N-2'
    FILING_N23C1 = 'n-23c-1'
    FILING_N23C2 = 'N-23C-2'
    FILING_N23C2_AMEND = 'N-23C-2/A'
    FILING_N23C3 = 'n-23c-3'
    FILING_N23C3A = 'N-23C3A'
    FILING_N23C3A_AMEND = 'N23C3A/A'
    FILING_N23C3B = 'N-23C3B'
    FILING_N23C3B_AMEND = 'N23C3B/A'
    FILING_N23C3C = 'N-23C3C'
    FILING_N23C3C_AMEND = 'N23C3C/A'
    FILING_N27D1 = 'N-27D-1'
    FILING_N27D1_AMEND = 'N-27D-1/A'
    FILING_N2MEF = 'N-2MEF'
    FILING_N2_AMEND = 'N-2/A'
    FILING_N3 = 'N-3'
    FILING_N30B2 = 'N-30B-2'
    FILING_N30D = 'N-30D'
    FILING_N30D_AMEND = 'N-30D/A'
    FILING_N3_AMEND = 'N-3/A'
    FILING_N3_EL = 'n-3 el'
    FILING_N4 = 'N-4'
    FILING_N4_AMEND = 'N-4/A'
    FILING_N4_EL = 'n-4 el'
    FILING_N5 = 'N-5'
    FILING_N54A = 'N-54A'
    FILING_N54A_AMEND = 'N-54A/A'
    FILING_N54C = 'N-54C'
    FILING_N54C_AMEND = 'N-54C/A'
    FILING_N5_AMEND = 'N-5/A'
    FILING_N6 = 'N-6'
    FILING_N6C9 = 'n-6c9'
    FILING_N6EI1 = 'n-6ei-1'
    FILING_N6F = 'N-6F'
    FILING_N6F_AMEND = 'N-6F/A'
    FILING_N6_AMEND = 'N-6/A'
    FILING_N8A = 'N-8A'
    FILING_N8A_AMEND = 'N-8A/A'
    FILING_N8B2 = 'N-8B-2'
    FILING_N8B2_AMEND = 'N-8B-2/A'
    FILING_N8B3 = 'N-8B-3'
    FILING_N8B3_AMEND = 'N-8B-3/A'
    FILING_N8B4 = 'N-8B-4'
    FILING_N8B4_AMEND = 'N-8B-4/A'
    FILING_N8F = 'N-8F'
    FILING_N8F_AMEND = 'N-8F/A'
    FILING_N8F_NTC = 'n-8f ntc'
    FILING_N8F_ORDR = 'n-8f ordr'
    FILING_NCEN = 'N-CEN'
    FILING_NCEN_AMEND = 'N-CEN/A'
    FILING_NCR = 'N-CR'
    FILING_NCR_AMEND = 'N-CR/A'
    FILING_NCSR = 'N-CSR'
    FILING_NCSRS = 'N-CSRS'
    FILING_NCSRS_AMEND = 'N-CSRS/A'
    FILING_NCSR_AMEND = 'N-CSR/A'
    FILING_NLIQUID = 'N-LIQUID'
    FILING_NLIQUID_AMEND = 'N-LIQUID/A'
    FILING_NMFP = 'n-mfp'
    FILING_NMFP2 = 'N-MFP2'
    FILING_NMFP2NMFP2_AMEND = 'N-MFP2 N-MFP2/A'
    FILING_NMFP2_AMEND = 'N-MFP2/A'
    FILING_NOINFOBLANKCELL = 'No info blank cell'
    FILING_NO_ACT = 'no act'
    FILING_NPORT = 'nport'
    FILING_NPORTEX = 'NPORT-EX'
    FILING_NPORTNP = 'NPORT-NP'
    FILING_NPORTNP_AMEND = 'NPORT-NP/A'
    FILING_NPORTP = 'NPORT-P'
    FILING_NPORTP_AMEND = 'NPORT-P/A'
    FILING_NPX = 'N-PX'
    FILING_NPXCR = 'N-PX-CR'
    FILING_NPXCR_AMEND = 'N-PX-CR/A'
    FILING_NPXFM = 'N-PX-FM'
    FILING_NPXFM_AMEND = 'N-PX-FM/A'
    FILING_NPXNT = 'N-PX-NT'
    FILING_NPXNT_AMEND = 'N-PX-NT/A'
    FILING_NPXVR = 'N-PX-VR'
    FILING_NPXVR_AMEND = 'N-PX-VR/A'
    FILING_NPX_AMEND = 'N-PX/A'
    FILING_NQ = 'N-Q'
    FILING_NQ_AMEND = 'N-Q/A'
    FILING_NRSRO = 'nrsro'
    FILING_NRSROCE = 'NRSRO-CE'
    FILING_NRSROCE_AMEND = 'NRSRO- CE/A'
    FILING_NRSROFR = 'NRSRO-FR'
    FILING_NRSROFR_AMEND = 'NRSRO-FR/A'
    FILING_NRSROUPD = 'NRSRO-UPD'
    FILING_NRSROWCLS = 'NRSRO-WCLS'
    FILING_NRSROWREG = 'NRSRO-WREG'
    FILING_NSARA = 'NSAR-A'
    FILING_NSARAT = 'nsar-at'
    FILING_NSARB = 'NSAR-B'
    FILING_NSARBT = 'nsar-bt'
    FILING_NSARB_AMEND = 'NSAR-B/A'
    FILING_NSARU = 'nsar-u'
    FILING_NT10D = 'NT 10-D'
    FILING_NT10D_AMEND = 'NT 10-D/A'
    FILING_NT10K = 'NT 10-K'
    FILING_NT10K_AMEND = 'NT 10-K/A'
    FILING_NT10Q = 'NT 10-Q'
    FILING_NT10Q_AMEND = 'NT 10-Q/A'
    FILING_NT11K = 'NT 11-K'
    FILING_NT11K_AMEND = 'NT 11-K/A'
    FILING_NT15D2 = 'NT 15D2'
    FILING_NT15D2_AMEND = 'NT 15D2/A'
    FILING_NT20F = 'NT 20-F'
    FILING_NT20F_AMEND = 'NT 20-F/A'
    FILING_NTFNCEN = 'ntfncen'
    FILING_NTFNCSR = 'ntfncsr'
    FILING_NTFNSAR = 'ntfnsar'
    FILING_NTN15D2 = 'ntn15d2'
    FILING_NTNCEN = 'NT-NCEN'
    FILING_NTNCEN_AMEND = 'NT-NCEN/A'
    FILING_NTNCSR = 'NT-NCSR'
    FILING_NTNCSR_AMEND = 'NTNCSR/A'
    FILING_NTNSAR = 'nt-nsar'
    FILING_NTN_10D = 'ntn 10-d'
    FILING_NTN_10K = 'ntn 10k'
    FILING_NTN_10Q = 'ntn 10q'
    FILING_NTN_11K = 'ntn 11k'
    FILING_NTN_20F = 'ntn 20f'
    FILING_NT_10D = 'nt 10-d'
    FILING_NT_10K = 'nt 10-k'
    FILING_NT_10Q = 'nt 10-q'
    FILING_NT_11K = 'nt 11-k'
    FILING_NT_15D2 = 'nt 15d2'
    FILING_NT_20F = 'nt 20-f'
    FILING_NT_NCEN = 'nt n-cen'
    FILING_NT_NMFP = 'nt n-mfp'
    FILING_NT_NPORTEX = 'nt nport-ex'
    FILING_NT_NPORTN = 'nt nport-n'
    FILING_NT_NPORTP = 'nt nport-p'
    FILING_NVP = 'N-VP'
    FILING_NVPFS = 'N-VPFS'
    FILING_NVPFS_AMEND = 'N-VPFS/A'
    FILING_NVP_AMEND = 'N-VP/A'
    FILING_OC = 'oc'
    FILING_OIP_NTC = 'oip ntc'
    FILING_OIP_ORDR = 'oip ordr'
    FILING_PF = 'pf'
    FILING_PILOT = 'pilot'
    FILING_POS462B = 'POS462B'
    FILING_POS462C = 'POS462C'
    FILING_POS8C = 'POS 8C'
    FILING_POSAM = 'POS AM'
    FILING_POSAMI = 'POS AMI'
    FILING_POSASR = 'POSASR'
    FILING_POSEX = 'POS EX'
    FILING_POS_8C = 'pos 8c'
    FILING_POS_AM = 'pos am'
    FILING_POS_AMC = 'pos amc'
    FILING_POS_AMI = 'pos ami'
    FILING_POS_EX = 'pos ex'
    FILING_PRE13E3 = 'pre13e3'
    FILING_PRE14A = 'PRE 14A'
    FILING_PRE14C = 'PRE 14C'
    FILING_PREA14A = 'prea14a'
    FILING_PREA14C = 'prea14c'
    FILING_PREC14A = 'PREC14A'
    FILING_PREC14C = 'PREC14C'
    FILING_PREM14A = 'PREM14A'
    FILING_PREM14C = 'PREM14C'
    FILING_PREN14A = 'PREN14A'
    FILING_PRER14A = 'PRER14A'
    FILING_PRER14C = 'PRER14C'
    FILING_PRES14A = 'pres14a'
    FILING_PRES14C = 'pres14c'
    FILING_PRE_14A = 'pre 14a'
    FILING_PRE_14C = 'pre 14c'
    FILING_PRRN14A = 'PRRN14A'
    FILING_PWRATT = 'pwr-att'
    FILING_PX14A6G = 'PX14A6G'
    FILING_PX14A6N = 'PX14A6N'
    FILING_QRTLYRPT = 'QRTLYRPT'
    FILING_QRTLYRPT_AMEND = 'QRTLYRPT/A'
    FILING_QUALIF = 'qualif'
    FILING_R31 = 'r31'
    FILING_REGDEX = 'regdex'
    FILING_REGNR = 'reg-nr'
    FILING_REVOKED = 'revoked'
    FILING_RW = 'RW'
    FILING_RWWD = 'RW WD'
    FILING_RW_WD = 'rw wd'
    FILING_S1 = 'S-1'
    FILING_S11 = 'S-11'
    FILING_S11MEF = 'S-11MEF'
    FILING_S11_AMEND = 'S-11/A'
    FILING_S1MEF = 'S-1MEF'
    FILING_S1_AMEND = 'S-1/A'
    FILING_S2 = 's-2'
    FILING_S20 = 'S-20'
    FILING_S20_AMEND = 'S-20/A'
    FILING_S2MEF = 's-2mef'
    FILING_S3 = 'S-3'
    FILING_S3ASR = 'S-3ASR'
    FILING_S3D = 'S-3D'
    FILING_S3DPOS = 'S-3DPOS'
    FILING_S3MEF = 'S-3MEF'
    FILING_S3_AMEND = 'S-3/A'
    FILING_S4 = 'S-4'
    FILING_S4EF = 'S-4EF'
    FILING_S4MEF = 'S-4MEF'
    FILING_S4POS = 'S-4 POS'
    FILING_S4_AMEND = 'S-4/A'
    FILING_S4_POS = 's-4 pos'
    FILING_S6 = 'S-6'
    FILING_S6EL24 = 's-6el24'
    FILING_S6_AMEND = 'S-6/A'
    FILING_S8 = 'S-8'
    FILING_S8POS = 'S-8 POS'
    FILING_S8_POS = 's-8 pos'
    FILING_SB = 'S-B'
    FILING_SB1 = 'sb-1'
    FILING_SB1MEF = 'sb-1mef'
    FILING_SB2 = 'sb-2'
    FILING_SB2MEF = 'sb-2mef'
    FILING_SBMEF = 'S-BMEF'
    FILING_SBSE = 'SBSE'
    FILING_SBSEA = 'SBSE-A'
    FILING_SBSEA_AMEND = 'SBSE-A/A'
    FILING_SBSEBD = 'SBSE-BD'
    FILING_SBSEBD_AMEND = 'SBSE-BD/A'
    FILING_SBSEC = 'SBSE-C'
    FILING_SBSEW = 'SBSE-W'
    FILING_SBSE_AMEND = 'SBSE/A'
    FILING_SB_AMEND = 'S-B/A'
    FILING_SC13D = 'SC 13D'
    FILING_SC13DA = 'sc 13d/a'
    FILING_SC13D_AMEND = 'SC 13D/A'
    FILING_SC13E1 = 'SC 13E1'
    FILING_SC13E1_AMEND = 'SC 13E1/A'
    FILING_SC13E3 = 'SC 13E3'
    FILING_SC13E3_AMEND = 'SC 13E3/A'
    FILING_SC13E4F = 'SC13E4F'
    FILING_SC13E4F_AMEND = 'SC13E4F/A'
    FILING_SC13G = 'SC 13G'
    FILING_SC13GA = 'sc 13g/a'
    FILING_SC13G_AMEND = 'SC 13G/A'
    FILING_SC14D1F = 'SC14D1F'
    FILING_SC14D1F_AMEND = 'SC14D1F/A'
    FILING_SC14D9 = 'SC 14D9'
    FILING_SC14D9C = 'SC14D9C'
    FILING_SC14D9F = 'SC14D9F'
    FILING_SC14D9F_AMEND = 'SC14D9F/A'
    FILING_SC14D9_AMEND = 'SC 14D9/A'
    FILING_SC14F1 = 'SC 14F1'
    FILING_SC14F1_AMEND = 'SC 14F1/A'
    FILING_SC14N = 'SC 14N'
    FILING_SC14NS = 'SC 14N-S'
    FILING_SC14NS_AMEND = 'SC 14N-S/A'
    FILING_SC14N_AMEND = 'SC 14N/A'
    FILING_SCI = 'sci'
    FILING_SCTOC = 'SC TO-C'
    FILING_SCTOI = 'SC TO-I'
    FILING_SCTOI_AMEND = 'SC TO-I/A'
    FILING_SCTOT = 'SC TO-T'
    FILING_SCTOT_AMEND = 'SC TO-T/A'
    FILING_SC_13D = 'sc 13d'
    FILING_SC_13E1 = 'sc 13e1'
    FILING_SC_13E3 = 'sc 13e3'
    FILING_SC_13E4 = 'sc 13e4'
    FILING_SC_13G = 'sc 13g'
    FILING_SC_14D1 = 'sc 14d1'
    FILING_SC_14D9 = 'sc 14d9'
    FILING_SC_14F1 = 'sc 14f1'
    FILING_SC_14N = 'sc 14n'
    FILING_SC_14NS = 'sc 14n-s'
    FILING_SC_TOC = 'sc to-c'
    FILING_SC_TOI = 'sc to-i'
    FILING_SC_TOT = 'sc to-t'
    FILING_SD = 'SD'
    FILING_SDR = 'SDR'
    FILING_SDRA = 'SDR-A'
    FILING_SDRCCO = 'SDR-CCO'
    FILING_SDRCCO_AMEND = 'SDR-CCO/A'
    FILING_SDRW = 'SDR-W'
    FILING_SDR_AMEND = 'SDR/A'
    FILING_SD_AMEND = 'SD/A'
    FILING_SE = 'se'
    FILING_SEC_ACTION = 'sec action'
    FILING_SEC_STAFF_ACTION = 'sec staff action'
    FILING_SEC_STAFF_LETTER = 'sec staff letter'
    FILING_SF1 = 'SF-1'
    FILING_SF1MEF = 'SF-1MEF'
    FILING_SF1_AMEND = 'SF-1/A'
    FILING_SF3 = 'SF-3'
    FILING_SF3MEF = 'SF-3MEF'
    FILING_SF3_AMEND = 'SF-3/A'
    FILING_SHER = 'SH-ER'
    FILING_SHER_AMEND = 'SH-ER/A'
    FILING_SHNT = 'SH-NT'
    FILING_SHNT_AMEND = 'SH-NT/A'
    FILING_SIP = 'sip'
    FILING_SL = 'sl'
    FILING_SP15D2 = 'SP 15D2'
    FILING_SP15D2_AMEND = 'SP 15D2/A'
    FILING_SP_15D2 = 'sp 15d2'
    FILING_STOP_ORDER = 'stop order'
    FILING_SUPPL = 'SUPPL'
    FILING_T1 = 'T-1'
    FILING_T2 = 'T-2'
    FILING_T3 = 'T-3'
    FILING_T3_AMEND = 'T-3/A'
    FILING_T4 = 't-4'
    FILING_T6 = 'T-6'
    FILING_T6_AMEND = 'T-6/A'
    FILING_TA1 = 'TA-1'
    FILING_TA1_AMEND = 'TA-1/A'
    FILING_TA2 = 'TA-2'
    FILING_TA2_AMEND = 'TA-2/A'
    FILING_TACO = 'taco'
    FILING_TAW = 'TA-W'
    FILING_TCR = 'tcr'
    FILING_TH = 'th'
    FILING_TTW = 'ttw'
    FILING_U1 = 'u-1'
    FILING_U12IA = 'u-12-ia'
    FILING_U12IB = 'u-12-ib'
    FILING_U131 = 'u-13-1'
    FILING_U1360 = 'u-13-60'
    FILING_U13E1 = 'u-13e-1'
    FILING_U33S = 'u-33-s'
    FILING_U3A2 = 'u-3a-2'
    FILING_U3A31 = 'u-3a3-1'
    FILING_U57 = 'u-57'
    FILING_U5A = 'u5a'
    FILING_U5B = 'u5b'
    FILING_U5S = 'u5s'
    FILING_U6B2 = 'u-6b-2'
    FILING_U7D = 'u-7d'
    FILING_U9C3 = 'u-9c-3'
    FILING_UNDER = 'UNDER'
    FILING_UNDER_AMEND = 'UNDER/A'
    FILING_UPLOAD = 'upload'
    FILING_UR1 = 'u-r-1'
    FILING_WBAPP = 'wb-app'
    FILING_WDLREQ = 'wdl-req'
    FILING_X17A5 = 'X-17A-5'
    FILING_X17A5_AMEND = 'X-17A-5/A'



from secedgar.core.combo import ComboFilings  # noqa:F401
from secedgar.core.company import CompanyFilings  # noqa:F401
from secedgar.core.daily import DailyFilings  # noqa:F401
from secedgar.core.filing_types import FilingType  # noqa:F401
from secedgar.core.quarterly import QuarterlyFilings  # noqa:F401



import os
from datetime import datetime

import click

from secedgar.core import CompanyFilings, DailyFilings, FilingType
from secedgar.exceptions import FilingTypeError


@click.group()
@click.option('-u', '--user-agent',
              help='Value used for HTTP header "User-Agent" for all requests.',
              required=True,
              type=str)
@click.pass_context
def cli(ctx, user_agent):
    """Main CLI group.

    Args:
        ctx (click.core.Context): Click context.
        user_agent (str): User agent string to pass.

    Returns:
        None
    """
    ctx.ensure_object(dict)
    ctx.obj['user_agent'] = user_agent


def date_cleanup(date):
    """Transforms date of form YYYYMMDD to datetime object.

    Args:
        date (str): Date of the form YYYYMMDD to be transformed.

    Returns:
        ``datetime.date`` object if given string.
        If given None, None is returned.
    """
    return datetime.strptime(date, "%Y%m%d").date() if date is not None else None


@cli.command()
@click.option('-l',
              '--lookups',
              help='Companies and tickers to include in filing download.',
              required=True,
              multiple=True)
@click.option('-t',
              '--ftype',
              help="""Choose a filing type.
             See ``secedgar.core.FilingType`` for a full list of available enums.
             Should be of the form FILING_<filing type>.""",
              required=True)  # Need to convert this to enum somehow
@click.option('-s',
              '--start',
              help="""Start date for filings.
              Should be in the format YYYYMMDD. Defaults to first available filing.""",
              type=str)
@click.option('-e', '--end',
              help='End date for filings. Should be in the format YYYYMMDD. Defaults to today.',
              type=str)
@click.option('-n', '--count',
              help='Number of filings to save. Defaults to all.',
              type=int)
@click.option('--directory',
              help="""Directory where files will be saved.
              Defaults to directory from which CLI is being executed.""",
              default=os.getcwd(), type=str)
@click.pass_context
def filing(ctx, lookups, ftype, start, end, count, directory):
    """Click command for downloading filings. Run ``secedgar filing --help`` for info.

    Args:
        ctx (click.core.Context): Click context.
        lookups (str): Companies and tickers to include in filing download.
        ftype (str): String of FilingType enum.
        start (str): Start date for filings in YYYYMMDD format.
            Will implicitly default to first available filing.
        end (str): End date for filings in YYYYMMDD format.
            Will implicitly default to today.
        count (int): Number of filings to save per ticker/company.
        directory (str): Directory where files should be saved.
            Defaults to current working directory.

    Returns:
        None
    """
    # If given filing type is not valid enum, raise FilingTypeError
    try:
        ftype = FilingType[ftype]
    except KeyError:
        raise FilingTypeError()

    f = CompanyFilings(cik_lookup=lookups,
                       filing_type=ftype,
                       start_date=date_cleanup(start),
                       end_date=date_cleanup(end),
                       count=count,
                       user_agent=ctx.obj['user_agent'])
    f.save(directory=directory)


@cli.command()
@click.option('-d',
              '--date',
              help="""Date to look up daily filings for.
              Should be in the format YYYYMMDD.""",
              required=True,
              type=str)
@click.option('--directory',
              help="""Directory where files will be saved.
              Defaults to directory from which CLI is being executed.""",
              default=os.getcwd(), type=str)
@click.pass_context
def daily(ctx, date, directory):
    """Click command for downloading daily filings. Run ``secedgar daily --help`` for info.

    Args:
        ctx (click.core.Context): Click context.
        date (str): Date to look up daily filings for. Should be in the format YYYYMMDD.
        directory (str): Directory where files should be saved.
            Defaults to current working directory.
    """
    d = DailyFilings(date=date_cleanup(date), user_agent=ctx.obj['user_agent'])
    d.save(directory=directory)



import datetime
import os


def sanitize_date(date):
    """Sanitizes date to be in acceptable format for EDGAR.

    Args:
        date (Union[datetime.datetime, datetime.date, str]): Date to be sanitized for request.

    Returns:
        date (str): Properly formatted date in 'YYYYMMDD' format.

    Raises:
        TypeError: If date is not in format YYYYMMDD as str or int.
    """
    if isinstance(date, (datetime.datetime, datetime.date)):
        return date.strftime("%Y%m%d")
    elif isinstance(date, str):
        if len(date) != 8:
            raise TypeError('Date must be of the form YYYYMMDD')
    elif isinstance(date, int):
        if date < 10 ** 7 or date > 10 ** 8:
            raise TypeError('Date must be of the form YYYYMMDD')
    return date


def make_path(path, **kwargs):
    """Make directory based on filing info.

    Args:
        path (str): Path to be made if it doesn't exist.
        **kwargs: Keyword arguments to pass to ``os.makedirs``.

    Raises:
        OSError: If there is a problem making the path.

    Returns:
        None
    """
    sanitized_path = os.path.expanduser(path)
    if not os.path.exists(sanitized_path):
        os.makedirs(sanitized_path, **kwargs)


def get_quarter(date):
    """Get quarter that corresponds with date.

    Args:
        date (Union[datetime.datetime, datetime.date]): Datetime object to get quarter for.
    """
    return (date.month - 1) // 3 + 1


def get_month(quarter):
    """Get month that corresponds with quarter start.

    Args:
        quarter (Union[datetime.datetime, datetime.date]): Datetime object to get quarter for.
    """
    if not isinstance(quarter, int):
        raise TypeError('Quarter must be an int')
    if quarter < 1 or quarter > 4:
        raise TypeError('Quarter must be between 1 and 4.')

    return 1 + (quarter - 1) * 3


def add_quarter(year, quarter):
    """Get the next year and quarter.

    Args:
        year (int): Year
        quarter (int): Quarter
    """
    if quarter < 1 or quarter > 4:
        raise TypeError('Quarter must be between 1 and 4.')
    if quarter == 4:
        quarter = 1
        year += 1
    else:
        quarter += 1
    return year, quarter



import json
from unittest.mock import patch

import pytest
import requests

from secedgar.cik_lookup import CIKLookup, get_cik_map
from secedgar.client import NetworkClient
from secedgar.exceptions import CIKError, EDGARQueryError
from secedgar.tests.conftest import MockResponse


@pytest.fixture
def ticker_lookups():
    return ["AAPL", "FB", "GOOGL", "NFLX", "MSFT"]


@pytest.fixture(scope="class")
def mock_client_cik_lookup(mock_user_agent):
    return NetworkClient(user_agent=mock_user_agent)


@pytest.fixture
def mock_single_cik_lookup_outside_map(monkeypatch):
    monkeypatch.setattr(NetworkClient, "get_response",
                        MockResponse(datapath_args=["CIK", "single_cik_search_result.html"]))


@pytest.fixture
def mock_get_cik_map(monkeypatch):
    response_json = {"0": {"cik_str": "320193", "ticker": "AAPL", "title": "Apple Inc."},
                     "1": {"cik_str": "789019", "ticker": "MSFT", "title": "MICROSOFT CORP"},
                     "2": {"cik_str": "1018724", "ticker": "AMZN", "title": "AMAZON COM INC"},
                     "3": {"cik_str": "1326801", "ticker": "FB", "title": "META PLATFORMS, INC."},
                     "4": {"cik_str": "1652044", "ticker": "GOOGL", "title": "Alphabet Inc."},
                     "5": {"cik_str": "1652044", "ticker": "GOOG", "title": "Alphabet Inc."}}
    response_json = json.dumps(response_json)
    monkeypatch.setattr(requests, 'get', MockResponse(content=bytes(response_json, "utf-8")))


@pytest.fixture
def mock_single_cik_lookup_response(monkeypatch):
    monkeypatch.setattr(NetworkClient, "get_response",
                        MockResponse(datapath_args=["CIK", "single_cik_search_result.html"]))


@pytest.fixture
def mock_single_cik_multiple_results_response(monkeypatch):
    monkeypatch.setattr(NetworkClient, "get_response",
                        MockResponse(datapath_args=["CIK", "cik_multiple_results.html"]))


@pytest.fixture
def mock_single_cik_not_found(monkeypatch):
    monkeypatch.setattr(NetworkClient, "get_response",
                        MockResponse(datapath_args=["CIK", "cik_not_found.html"]))


class TestCIKLookup(object):
    ciklookup_results = [
        (['AAPL'], {'AAPL': '320193'}),
        (['AAPL', 'AMAZON COM INC'], {'AAPL': '320193', 'AMAZON COM INC': '1018724'}),
        (['Alphabet Inc.'], {'Alphabet Inc.': '1652044'}),
        (['aapl', 'msft'], {'aapl': '320193', 'msft': '789019'}),
        (['aapl', 'msft', 'Alphabet Inc.'],
         {'aapl': '320193', 'msft': '789019', 'Alphabet Inc.': '1652044'}),
        (['320193'], {'320193': '320193'}),
        (['320193', '1018724'], {'320193': '320193', '1018724': '1018724'}),
        (['AAPL', '1018724'], {'AAPL': '320193', '1018724': '1018724'}),
    ]
    cikmap_ticker_results = [
        ("aapl", "320193"),
        ("AMZN", "1018724"),
        ("Msft", "789019"),
    ]
    cikmap_company_results = [
        ("meta platforms, inc.", "1326801"),
        ("Alphabet Inc.", "1652044"),
    ]

    @pytest.mark.parametrize(
        "lookup,expected",
        ciklookup_results
    )
    def test_cik_lookup_returns_correct_values(self, mock_client_cik_lookup,
                                               lookup,
                                               expected,
                                               mock_get_cik_map):
        look = CIKLookup(lookup, client=mock_client_cik_lookup)
        assert look.lookup_dict == expected

    @pytest.mark.smoke
    @pytest.mark.parametrize(
        "lookup,expected",
        ciklookup_results
    )
    def test_cik_lookup_returns_correct_values_smoke(self, lookup,
                                                     expected,
                                                     real_test_client):
        look = CIKLookup(lookup, client=real_test_client)
        assert look.lookup_dict == expected

    def test_cik_lookup_lookups_property(self, mock_client_cik_lookup):
        multiple_company_lookup = CIKLookup(['aapl', 'msft', 'fb'], client=mock_client_cik_lookup)
        assert multiple_company_lookup.lookups == ['aapl', 'msft', 'fb']

    def test_multiple_results_company_name_search(self, mock_client_cik_lookup,
                                                  mock_single_cik_multiple_results_response):
        multiple_results_cik = CIKLookup('paper', client=mock_client_cik_lookup)
        with pytest.warns(UserWarning):
            assert len(multiple_results_cik.ciks) == 0

    def test_multiple_results_raises_warnings(self, mock_client_cik_lookup,
                                              mock_single_cik_multiple_results_response):
        multiple_results_cik = CIKLookup('paper', client=mock_client_cik_lookup)
        with pytest.warns(UserWarning):
            _ = multiple_results_cik.ciks

    def test_cik_lookup_cik_hits_request(self, mock_client_cik_lookup):
        with patch.object(CIKLookup, '_get_cik_from_html') as mock:
            CIKLookup(['Apple'], client=mock_client_cik_lookup).get_ciks()
            mock.assert_called()

    def test_cik_lookup_cik_bypasses_request(self, mock_client_cik_lookup):
        with patch.object(CIKLookup, '_get_cik_from_html') as mock:
            CIKLookup(['1018724'], client=mock_client_cik_lookup).get_ciks()
            mock.assert_not_called()

    @pytest.mark.parametrize(
        "bad_cik",
        [
            1234567890,
            123.0,
        ]
    )
    def test_validate_cik__is_string(self, mock_client_cik_lookup, bad_cik):
        with pytest.raises(TypeError):
            CIKLookup(bad_cik, client=mock_client_cik_lookup)

    def test_validate_cik_after_cik_lookup(self, mock_client_cik_lookup,
                                           mock_single_cik_not_found):
        # string remains unchecked until query to allow for possibility of
        # using company name, ticker, or CIK as string
        with pytest.raises(EDGARQueryError):
            _ = CIKLookup('0notvalid0', client=mock_client_cik_lookup).ciks

    def test_client_property(self, mock_client_cik_lookup, ticker_lookups):
        lookup = CIKLookup(ticker_lookups, client=mock_client_cik_lookup)
        assert lookup.client == mock_client_cik_lookup

    @pytest.mark.parametrize(
        "bad_lookups",
        [
            [],
            "",
            ["AAPL", 4, "FB"]
        ]
    )
    def test_bad_lookups_raises_type_error(self, bad_lookups, mock_client_cik_lookup):
        with pytest.raises(TypeError):
            CIKLookup(lookups=bad_lookups, client=mock_client_cik_lookup)

    def test_lookups_property(self, mock_client_cik_lookup, ticker_lookups):
        lookup = CIKLookup(lookups=ticker_lookups, client=mock_client_cik_lookup)
        assert lookup.lookups == ticker_lookups

    @pytest.mark.parametrize(
        "bad_lookup",
        [
            "",
            4
        ]
    )
    def test_validate_lookup(self, bad_lookup):
        with pytest.raises(TypeError):
            CIKLookup._validate_lookup(bad_lookup)

    @pytest.mark.parametrize(
        "bad_cik",
        [
            "1234",
            1234,
            1234567890,
            "AAPL",
            "",
            None,
        ]
    )
    def test_validate_cik_on_bad_ciks(self, bad_cik):
        with pytest.raises(CIKError):
            CIKLookup._validate_cik(bad_cik)

    def test_params_reset_after_get_cik(self, ticker_lookups, mock_client_cik_lookup,
                                        mock_single_cik_lookup_outside_map):
        lookup = CIKLookup(lookups=ticker_lookups, client=mock_client_cik_lookup)
        lookup._get_cik_from_html(ticker_lookups[0])
        assert lookup.params.get("CIK") is None and lookup.params.get("company") is None

    @pytest.mark.parametrize(
        "lookup,cik",
        cikmap_ticker_results
    )
    def test_get_cik_map_tickers(self, lookup, cik, mock_get_cik_map, mock_user_agent):
        cik_map = get_cik_map(user_agent=mock_user_agent)
        assert cik_map["ticker"][lookup.upper()] == cik

    @pytest.mark.smoke
    @pytest.mark.parametrize(
        "lookup,cik",
        cikmap_ticker_results
    )
    def test_get_cik_map_tickers_smoke(self, lookup, cik, mock_user_agent):
        cik_map = get_cik_map(user_agent=mock_user_agent)
        assert cik_map["ticker"][lookup.upper()] == cik

    @pytest.mark.parametrize(
        "lookup,cik",
        cikmap_company_results
    )
    def test_get_cik_map_company_names(self, lookup, cik, mock_get_cik_map, mock_user_agent):
        cik_map = get_cik_map(user_agent=mock_user_agent)
        assert cik_map["title"][lookup.upper()] == cik

    @pytest.mark.smoke
    @pytest.mark.parametrize(
        "lookup,cik",
        cikmap_company_results
    )
    def test_get_cik_map_company_names_smoke(self, lookup, cik, mock_user_agent):
        cik_map = get_cik_map(user_agent=mock_user_agent)
        assert cik_map["title"][lookup.upper()] == cik

    @pytest.mark.smoke
    def test_get_cik_map_smoke(self, mock_user_agent):
        cik_map = get_cik_map(user_agent=mock_user_agent)
        assert "ticker" in cik_map and "title" in cik_map
        assert cik_map["ticker"] and cik_map["title"]

    def test_cik_lookup_client_none(self, ticker_lookups, mock_user_agent):
        lookup = CIKLookup(lookups=ticker_lookups, user_agent=mock_user_agent, client=None)
        assert isinstance(lookup.client, NetworkClient)
        assert lookup.lookups == ticker_lookups
